# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=cyclic-import
"""The AiiDA process class"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import collections
import enum
import inspect
import uuid
import traceback

import six
from six.moves import filter, range
from pika.exceptions import ConnectionClosed

import plumpy
from plumpy import ProcessState

from aiida.common import exceptions
from aiida.common.extendeddicts import AttributeDict
from aiida.common.lang import classproperty, override, protected
from aiida.common.links import LinkType
from aiida.common.log import LOG_LEVEL_REPORT
from aiida import orm
from aiida.orm import ProcessNode, CalculationNode, WorkflowNode
from aiida.orm.utils import serialize

from .. import utils
from .exit_code import ExitCode
from .builder import ProcessBuilder
from .ports import InputPort, PortNamespace
from .process_spec import ProcessSpec

__all__ = ('Process', 'ProcessState')


def instantiate_process(runner, process, *args, **inputs):
    """
    Return an instance of the process with the given inputs. The function can deal with various types
    of the `process`:

        * Process instance: will simply return the instance
        * ProcessBuilder instance: will instantiate the Process from the class and inputs defined within it
        * Process class: will instantiate with the specified inputs

    If anything else is passed, a ValueError will be raised

    :param process: Process instance or class, CalcJobNode class or ProcessBuilder instance
    :param inputs: the inputs for the process to be instantiated with
    """
    if isinstance(process, Process):
        assert not args
        assert not inputs
        assert runner is process.runner
        return process

    if isinstance(process, ProcessBuilder):
        builder = process
        process_class = builder.process_class
        inputs.update(**builder)
    elif issubclass(process, Process):
        process_class = process
    else:
        raise ValueError('invalid process {}, needs to be Process or ProcessBuilder'.format(type(process)))

    process = process_class(runner=runner, inputs=inputs)

    return process


@plumpy.auto_persist('_parent_pid', '_enable_persistence')
@six.add_metaclass(abc.ABCMeta)
class Process(plumpy.Process):
    """
    This class represents an AiiDA process which can be executed and will
    have full provenance saved in the database.
    """
    # pylint: disable=too-many-public-methods

    _node_class = ProcessNode
    _spec_type = ProcessSpec

    SINGLE_OUTPUT_LINKNAME = 'result'

    class SaveKeys(enum.Enum):
        """
        Keys used to identify things in the saved instance state bundle.
        """
        # pylint: disable=too-few-public-methods
        CALC_ID = 'calc_id'

    @classmethod
    def define(cls, spec):
        super(Process, cls).define(spec)
        spec.input_namespace(spec.metadata_key, required=False, non_db=True, default={})
        spec.input_namespace('{}.{}'.format(spec.metadata_key, spec.options_key), required=False)
        spec.input('{}.store_provenance'.format(spec.metadata_key), valid_type=bool, default=True, non_db=True)
        spec.input(
            '{}.description'.format(spec.metadata_key), valid_type=six.string_types[0], required=False, non_db=True)
        spec.input('{}.label'.format(spec.metadata_key), valid_type=six.string_types[0], required=False, non_db=True)
        spec.inputs.valid_type = (orm.Data,)
        spec.outputs.valid_type = (orm.Data,)

    @classmethod
    def get_builder(cls):
        return ProcessBuilder(cls)

    @classmethod
    def get_or_create_db_record(cls):
        """
        Create a database calculation node that represents what happened in
        this process.
        :return: A calculation
        """
        return cls._node_class()

    def __init__(self, inputs=None, logger=None, runner=None, parent_pid=None, enable_persistence=True):
        from aiida.manage import manager

        self._runner = runner if runner is not None else manager.get_manager().get_runner()

        super(Process, self).__init__(
            inputs=self.spec().inputs.serialize(inputs),
            logger=logger,
            loop=self._runner.loop,
            communicator=self.runner.communicator)

        self._node = None
        self._parent_pid = parent_pid
        self._enable_persistence = enable_persistence
        if self._enable_persistence and self.runner.persister is None:
            self.logger.warning('Disabling persistence, runner does not have a persister')
            self._enable_persistence = False

    def init(self):
        super(Process, self).init()
        if self._logger is None:
            self.set_logger(self.node.logger)

    @classproperty
    def exit_codes(self):
        """
        Return the namespace of exit codes defined for this WorkChain through its ProcessSpec.
        The namespace supports getitem and getattr operations with an ExitCode label to retrieve a specific code.
        Additionally, the namespace can also be called with either the exit code integer status to retrieve it.

        :returns: ExitCodesNamespace of ExitCode named tuples
        """
        return self.spec().exit_codes

    @property
    def node(self):
        """Return the ProcessNode used by this process to represent itself in the database.

        :return: instance of sub class of ProcessNode
        """
        return self._node

    @property
    def uuid(self):
        """Return the UUID of the process which corresponds to the UUID of its associated `ProcessNode`.

        :return: the UUID associated to this process instance
        """
        return self.node.uuid

    @property
    def metadata(self):
        """Return the metadata passed when launching this process.

        :return: metadata dictionary
        """
        try:
            return self.inputs.metadata
        except AttributeError:
            return AttributeDict()

    @property
    def options(self):
        """Return the options of the metadata passed when launching this process.

        :return: options dictionary
        """
        try:
            return self.metadata.options
        except AttributeError:
            return AttributeDict()

    def _save_checkpoint(self):
        """
        Save the current state in a chechpoint if persistence is enabled and the process state is not terminal

        If the persistence call excepts with a PersistenceError, it will be caught and a warning will be logged.
        """
        if self._enable_persistence and not self._state.is_terminal():
            try:
                self.runner.persister.save_checkpoint(self)
            except plumpy.PersistenceError:
                self.logger.exception("Exception trying to save checkpoint, this means you will "
                                      "not be able to restart in case of a crash until the next successful checkpoint.")

    @override
    def save_instance_state(self, out_state, save_context):
        super(Process, self).save_instance_state(out_state, save_context)

        if self.metadata.store_provenance:
            assert self.node.is_stored

        out_state[self.SaveKeys.CALC_ID.value] = self.pid

    def get_provenance_inputs_iterator(self):
        return filter(lambda kv: not kv[0].startswith('_'), self.inputs.items())

    @override
    def load_instance_state(self, saved_state, load_context):
        from aiida.manage import manager

        if 'runner' in load_context:
            self._runner = load_context.runner
        else:
            self._runner = manager.get_manager().get_runner()

        load_context = load_context.copyextend(loop=self._runner.loop, communicator=self._runner.communicator)
        super(Process, self).load_instance_state(saved_state, load_context)

        if self.SaveKeys.CALC_ID.value in saved_state:
            self._node = orm.load_node(saved_state[self.SaveKeys.CALC_ID.value])
            self._pid = self.node.pk
        else:
            self._pid = self._create_and_setup_db_record()

        self.node.logger.info('Loaded process<{}> from saved state'.format(self.node.pk))

    def kill(self, msg=None):
        """
        Kill the process and all the children calculations it called
        """
        self.node.logger.debug('Request to kill Process<{}>'.format(self.node.pk))

        had_been_terminated = self.has_terminated()

        result = super(Process, self).kill(msg)

        # Only kill children if we could be killed ourselves
        if result is not False and not had_been_terminated:
            killing = []
            for child in self.node.called:
                try:
                    result = self.runner.controller.kill_process(child.pk, 'Killed by parent<{}>'.format(self.node.pk))
                    if isinstance(result, plumpy.Future):
                        killing.append(result)
                except ConnectionClosed:
                    self.logger.info('no connection available to kill child<%s>', child.pk)

            if isinstance(result, plumpy.Future):
                # We ourselves are waiting to be killed so add it to the list
                killing.append(result)

            if killing:
                # We are waiting for things to be killed, so return the 'gathered' future
                result = plumpy.gather(result)

        return result

    @override
    def out(self, output_port, value=None):
        if value is None:
            # In this case assume that output_port is the actual value and there is just one return value
            value = output_port
            output_port = self.SINGLE_OUTPUT_LINKNAME

        return super(Process, self).out(output_port, value)

    def out_many(self, out_dict):
        """
        Add all values given in ``out_dict`` to the outputs. The keys of the dictionary will be used as output names.
        """
        for key, value in out_dict.items():
            self.out(key, value)

    # region Process event hooks
    def on_create(self):
        super(Process, self).on_create()
        # If parent PID hasn't been supplied try to get it from the stack
        if self._parent_pid is None and Process.current():
            current = Process.current()
            if isinstance(current, Process):
                self._parent_pid = current.pid
        self._pid = self._create_and_setup_db_record()

    @override
    def on_entering(self, state):
        super(Process, self).on_entering(state)
        # Update the node attributes every time we enter a new state

    def on_entered(self, from_state):
        self.update_node_state(self._state)
        self._save_checkpoint()
        # Update the latest process state change timestamp
        utils.set_process_state_change_timestamp(self)
        super(Process, self).on_entered(from_state)

    @override
    def on_terminated(self):
        """
        Called when a Process enters a terminal state.
        """
        super(Process, self).on_terminated()
        if self._enable_persistence:
            try:
                self.runner.persister.delete_checkpoint(self.pid)
            except BaseException:
                self.logger.exception('Failed to delete checkpoint')

        try:
            self.node.seal()
        except exceptions.ModificationNotAllowed:
            pass

    @override
    def on_except(self, exc_info):
        """
        Log the exception by calling the report method with formatted stack trace from exception info object
        and store the exception string as a node attribute

        :param exc_info: the sys.exc_info() object
        """
        super(Process, self).on_except(exc_info)
        self.node.set_exception(''.join(traceback.format_exception(exc_info[0], exc_info[1], None)))
        self.report(''.join(traceback.format_exception(*exc_info)))

    @override
    def on_finish(self, result, successful):
        """
        Set the finish status on the process node
        """
        super(Process, self).on_finish(result, successful)

        if result is None or isinstance(result, int):
            self.node.set_exit_status(result)
        elif isinstance(result, ExitCode):
            self.node.set_exit_status(result.status)
            self.node.set_exit_message(result.message)
        else:
            raise ValueError('the result should be an integer, ExitCode or None, got {} {} {}'.format(
                type(result), result, self.pid))

    @override
    def on_paused(self, msg=None):
        """
        The Process was paused so set the paused attribute on the process node
        """
        super(Process, self).on_paused(msg)
        self._save_checkpoint()
        self.node.pause()

    @override
    def on_playing(self):
        """
        The Process was unpaused so remove the paused attribute on the process node
        """
        super(Process, self).on_playing()
        self.node.unpause()

    @override
    def on_output_emitting(self, output_port, value):
        """
        The process has emitted a value on the given output port.

        :param output_port: The output port name the value was emitted on
        :param value: The value emitted
        """
        super(Process, self).on_output_emitting(output_port, value)

        if not isinstance(value, orm.Data):
            raise TypeError('Values output from process must be instances of AiiDA orm.Data types, got {}'.format(
                value.__class__))

    # end region

    def set_status(self, status):
        """
        The status of the Process is about to be changed, so we reflect this is in node's attribute proxy.

        :param status: the status message
        """
        super(Process, self).set_status(status)
        self.node.set_process_status(status)

    def submit(self, process, *args, **kwargs):
        return self.runner.submit(process, *args, **kwargs)

    @property
    def runner(self):
        return self._runner

    @protected
    def get_parent_calc(self):
        """
        Get the parent process node

        :return: the parent process node if there is one
        :rtype: :class:`aiida.orm.nodes.process.process.ProcessNode`
        """
        # Can't get it if we don't know our parent
        if self._parent_pid is None:
            return None

        return orm.load_node(pk=self._parent_pid)

    @classmethod
    def build_process_type(cls):
        """
        The process type.

        :return: string of the process type
        :rtype: str

        Note: This could be made into a property 'process_type' but in order to have it be a property of the class
        it would need to be defined in the metaclass, see https://bugs.python.org/issue20659
        """
        from aiida.plugins.entry_point import get_entry_point_string_from_class

        class_module = cls.__module__
        class_name = cls.__name__

        # If the process is a registered plugin the corresponding entry point will be used as process type
        process_type = get_entry_point_string_from_class(class_module, class_name)

        # If no entry point was found, default to fully qualified path name
        if process_type is None:
            return '{}.{}'.format(class_module, class_name)

        return process_type

    @protected
    def report(self, msg, *args, **kwargs):
        """Log a message to the logger, which should get saved to the database through the attached DbLogHandler.

        The pk, class name and function name of the caller are prepended to the given message

        :param msg: message to log
        :param args: args to pass to the log call
        :param kwargs: kwargs to pass to the log call
        """
        message = '[{}|{}|{}]: {}'.format(self.node.pk, self.__class__.__name__, inspect.stack()[1][3], msg)
        self.logger.log(LOG_LEVEL_REPORT, message, *args, **kwargs)

    def _create_and_setup_db_record(self):
        """
        Create and setup the database record for this process

        :return: the uuid of the process
        """
        self._node = self.get_or_create_db_record()
        self._setup_db_record()
        if self.metadata.store_provenance:
            try:
                self.node.store_all()
                if self.node.is_finished_ok:
                    self._state = ProcessState.FINISHED
                    for entry in self.node.get_outgoing(link_type=LinkType.RETURN):
                        if entry.link_label.endswith('_{pk}'.format(pk=entry.node.pk)):
                            continue
                        self.out(entry.link_label, entry.node)
                    # This is needed for CalcJob. In that case, the outputs are
                    # returned regardless of whether they end in '_pk'
                    for entry in self.node.get_outgoing(link_type=LinkType.CREATE):
                        self.out(entry.link_label, entry.node)
            except exceptions.ModificationNotAllowed:
                # The calculation was already stored
                pass

        if self.node.pk is not None:
            return self.node.pk

        return uuid.UUID(self.node.uuid)

    @override
    def encode_input_args(self, inputs):
        """
        Encode input arguments such that they may be saved in a Bundle

        :param inputs: A mapping of the inputs as passed to the process
        :return: The encoded inputs
        """
        return serialize.serialize(inputs)

    @override
    def decode_input_args(self, encoded):
        """
        Decode saved input arguments as they came from the saved instance state Bundle

        :param encoded:
        :return: The decoded input args
        """
        return serialize.deserialize(encoded)

    def update_node_state(self, state):
        self.update_outputs()
        self.node.set_process_state(state.LABEL)

    def update_outputs(self):
        """Attach any new outputs to the node since the last time this was called, if store provenance is True."""
        if self.metadata.store_provenance is False:
            return

        outputs_stored = self.node.get_outgoing(link_type=(LinkType.CREATE, LinkType.RETURN)).all_link_labels()
        outputs_new = set(self.outputs.keys()) - set(outputs_stored)

        for link_label in outputs_new:

            output = self.outputs[link_label]

            if isinstance(self.node, CalculationNode):
                output.add_incoming(self.node, LinkType.CREATE, link_label)
            elif isinstance(self.node, WorkflowNode):
                output.add_incoming(self.node, LinkType.RETURN, link_label)

            output.store()

    def _setup_db_record(self):
        """
        Create the database record for this process and the links with respect to its inputs

        This function will set various attributes on the node that serve as a proxy for attributes of the Process.
        This is essential as otherwise this information could only be introspected through the Process itself, which
        is only available to the interpreter that has it in memory. To make this data introspectable from any
        interpreter, for example for the command line interface, certain Process attributes are proxied through the
        calculation node.

        In addition, the parent calculation will be setup with a CALL link if applicable and all inputs will be
        linked up as well.
        """
        assert self.inputs is not None
        assert not self.node.is_sealed, 'process node cannot be sealed when setting up the database record'

        # Store important process attributes in the node proxy
        self.node.set_process_state(None)
        self.node.set_process_label(self.__class__.__name__)
        self.node.set_process_type(self.__class__.build_process_type())

        parent_calc = self.get_parent_calc()

        if parent_calc and self.metadata.store_provenance:

            if isinstance(parent_calc, CalculationNode):
                raise exceptions.InvalidOperation('calling processes from a calculation type process is forbidden.')

            if isinstance(self.node, CalculationNode):
                self.node.add_incoming(parent_calc, LinkType.CALL_CALC, 'CALL_CALC')

            elif isinstance(self.node, WorkflowNode):
                self.node.add_incoming(parent_calc, LinkType.CALL_WORK, 'CALL_WORK')

        self._setup_metadata()
        self._setup_inputs()

    def _setup_metadata(self):
        """Store the metadata on the ProcessNode."""
        for name, metadata in self.metadata.items():
            if name == 'store_provenance':
                continue
            elif name == 'label':
                self.node.label = metadata
            elif name == 'description':
                self.node.description = metadata
            elif name == 'options':
                for option_name, option_value in metadata.items():
                    self.node.set_option(option_name, option_value)
            else:
                raise RuntimeError('unsupported metadata key: {}'.format(name))

    def _setup_inputs(self):
        """Create the links between the input nodes and the ProcessNode that represents this process."""
        from aiida.orm import Code

        for name, node in self._flat_inputs().items():

            # Certain processes allow to specify ports with `None` as acceptable values
            if node is None:
                continue

            # Special exception: set computer if node is a remote Code and our node does not yet have a computer set
            if isinstance(node, Code) and not node.is_local() and not self.node.computer:
                self.node.computer = node.get_remote_computer()

            # Need this special case for tests that use ProcessNodes as classes
            if isinstance(self.node, ProcessNode) and not isinstance(self.node, (CalculationNode, WorkflowNode)):
                self.node.add_incoming(node, LinkType.INPUT_WORK, name)

            elif isinstance(self.node, CalculationNode):
                self.node.add_incoming(node, LinkType.INPUT_CALC, name)

            elif isinstance(self.node, WorkflowNode):
                self.node.add_incoming(node, LinkType.INPUT_WORK, name)

    def _flat_inputs(self):
        """
        Return a flattened version of the parsed inputs dictionary.

        The eventual keys will be a concatenation of the nested keys. Note that the `metadata` dictionary, if present,
        is not passed, as those are dealt with separately in `_setup_metadata`.

        :return: flat dictionary of parsed inputs
        """
        inputs = {key: value for key, value in self.inputs.items() if key != self.spec().metadata_key}
        return dict(self._flatten_inputs(self.spec().inputs, inputs))

    def _flatten_inputs(self, port, port_value, parent_name='', separator='_'):
        """
        Function that will recursively flatten the inputs dictionary, omitting inputs for ports that
        are marked as being non database storable

        :param port: port against which to map the port value, can be InputPort or PortNamespace
        :param port_value: value for the current port, can be a Mapping
        :param parent_name: the parent key with which to prefix the keys
        :param separator: character to use for the concatenation of keys
        """
        if ((port is None and isinstance(port_value, orm.Node)) or
            (isinstance(port, InputPort) and not getattr(port, 'non_db', False))):
            return [(parent_name, port_value)]

        if (port is None and isinstance(port_value, collections.Mapping) or isinstance(port, PortNamespace)):
            items = []
            for name, value in port_value.items():

                prefixed_key = parent_name + separator + name if parent_name else name

                try:
                    nested_port = port[name]
                except (KeyError, TypeError):
                    nested_port = None

                sub_items = self._flatten_inputs(
                    port=nested_port, port_value=value, parent_name=prefixed_key, separator=separator)
                items.extend(sub_items)
            return items

        assert (port is None) or (isinstance(port, InputPort) and port.non_db)
        return []

    def exposed_inputs(self, process_class, namespace=None, agglomerate=True):
        """
        Gather a dictionary of the inputs that were exposed for a given Process class under an optional namespace.

        :param process_class: Process class whose inputs to try and retrieve

        :param namespace: PortNamespace in which to look for the inputs
        :type namespace: str

        :param agglomerate: If set to true, all parent namespaces of the given ``namespace`` will also be
            searched for inputs. Inputs in lower-lying namespaces take precedence.
        :type agglomerate: bool
        """
        exposed_inputs = {}

        namespace_list = self._get_namespace_list(namespace=namespace, agglomerate=agglomerate)
        for nspace in namespace_list:

            # The nspace None indicates the base level nspace
            if nspace is None:
                inputs = self.inputs
                port_nspace = self.spec().inputs
            else:
                inputs = self.inputs
                for part in nspace.split('.'):
                    inputs = inputs[part]
                try:
                    port_nspace = self.spec().inputs.get_port(nspace)
                except KeyError:
                    raise ValueError('this process does not contain the "{}" input nspace'.format(nspace))

            # Get the list of ports that were exposed for the given Process class in the current nspace
            exposed_inputs_list = self.spec()._exposed_inputs[nspace][process_class]  # pylint: disable=protected-access

            for name in port_nspace.ports.keys():
                if name in inputs and name in exposed_inputs_list:
                    exposed_inputs[name] = inputs[name]

        return exposed_inputs

    def exposed_outputs(self, process_instance, process_class, namespace=None, agglomerate=True):
        """
        Gather the outputs which were exposed from the ``process_class`` and emitted by the specific
        ``process_instance`` in a dictionary.

        :param namespace: Namespace in which to search for exposed outputs.
        :type namespace: str

        :param agglomerate: If set to true, all parent namespaces of the given ``namespace`` will also
            be searched for outputs. Outputs in lower-lying namespaces take precedence.
        :type agglomerate: bool
        """
        namespace_separator = self.spec().namespace_separator

        output_key_map = {}
        # maps the exposed name to all outputs that belong to it
        top_namespace_map = collections.defaultdict(list)
        process_outputs_dict = {
            entry.link_label: entry.node for entry in process_instance.get_outgoing(link_type=LinkType.RETURN)
        }

        for port_name in process_outputs_dict:
            top_namespace = port_name.split(namespace_separator)[0]
            top_namespace_map[top_namespace].append(port_name)

        for nspace in self._get_namespace_list(namespace=namespace, agglomerate=agglomerate):
            # only the top-level key is stored in _exposed_outputs
            for top_name in top_namespace_map:
                if top_name in self.spec()._exposed_outputs[nspace][process_class]:  # pylint: disable=protected-access
                    output_key_map[top_name] = nspace

        result = {}

        for top_name, nspace in output_key_map.items():
            # collect all outputs belonging to the given top_name
            for port_name in top_namespace_map[top_name]:
                if nspace is None:
                    result[port_name] = process_outputs_dict[port_name]
                else:
                    result[nspace + namespace_separator + port_name] = process_outputs_dict[port_name]
        return result

    @staticmethod
    def _get_namespace_list(namespace=None, agglomerate=True):
        """Get the list of namespaces in a given namespace"""
        if not agglomerate:
            return [namespace]

        namespace_list = [None]
        if namespace is not None:
            split_ns = namespace.split('.')
            namespace_list.extend(['.'.join(split_ns[:i]) for i in range(1, len(split_ns) + 1)])
        return namespace_list


def get_query_string_from_process_type_string(process_type_string):  # pylint: disable=invalid-name
    """
    Take the process type string of a Node and create the queryable type string.

    :param process_type_string: the process type string
    :return: string that can be used to query for subclasses of the process type using 'LIKE <string>'
    """
    if ":" in process_type_string:
        return process_type_string + "."

    path = process_type_string.rsplit('.', 2)[0]
    return path + "."



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Various utils that should be used during migrations and migrations tests because the AiiDA ORM cannot be used."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import errno
import io
import os
import numpy


def ensure_repository_folder_created(uuid):
    """Make sure that the repository sub folder for the node with the given UUID exists or create it.

    :param uuid: UUID of the node
    """
    dirpath = get_node_repository_sub_folder(uuid)

    try:
        os.makedirs(dirpath)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def put_object_from_string(uuid, name, content):
    """Write a file with the given content in the repository sub folder of the given node.

    :param uuid: UUID of the node
    :param name: name to use for the file
    :param content: the content to write to the file
    """
    ensure_repository_folder_created(uuid)
    filepath = os.path.join(get_node_repository_sub_folder(uuid), name)

    with io.open(filepath, 'w', encoding='utf-8') as handle:
        handle.write(content)


def get_object_from_repository(uuid, name):
    """Return the content of a file with the given name in the repository sub folder of the given node.

    :param uuid: UUID of the node
    :param name: name to use for the file
    """
    filepath = os.path.join(get_node_repository_sub_folder(uuid), name)

    with io.open(filepath) as handle:
        return handle.read()


def get_node_repository_sub_folder(uuid):
    """Return the absolute path to the sub folder `path` within the repository of the node with the given UUID.

    :param uuid: UUID of the node
    :return: absolute path to node repository folder, i.e `/some/path/repository/node/12/ab/c123134-a123/path`
    """
    from aiida.common.utils import get_repository_folder

    uuid = str(uuid)

    repo_dirpath = get_repository_folder('repository')
    node_dirpath = os.path.join(repo_dirpath, 'node', uuid[:2], uuid[2:4], uuid[4:], 'path')

    return node_dirpath


def get_numpy_array_absolute_path(uuid, name):
    """Return the absolute path of a numpy array with the given name in the repository of the node with the given uuid.

    :param uuid: the UUID of the node
    :param name: the name of the numpy array
    :return: the absolute path of the numpy array file
    """
    return os.path.join(get_node_repository_sub_folder(uuid), name + '.npy')


def store_numpy_array_in_repository(uuid, name, array):
    """Store a numpy array in the repository folder of a node.

    :param uuid: the node UUID
    :param name: the name under which to store the array
    :param array: the numpy array to store
    """
    ensure_repository_folder_created(uuid)
    filepath = get_numpy_array_absolute_path(uuid, name)

    with io.open(filepath, 'wb') as handle:
        numpy.save(handle, array)


def delete_numpy_array_from_repository(uuid, name):
    """Delete the numpy array with a given name from the repository corresponding to a node with a given uuid.

    :param uuid: the UUID of the node
    :param name: the name of the numpy array
    """
    filepath = get_numpy_array_absolute_path(uuid, name)

    try:
        os.remove(filepath)
    except (IOError, OSError):
        pass


def load_numpy_array_from_repository(uuid, name):
    """Load and return a numpy array from the repository folder of a node.

    :param uuid: the node UUID
    :param name: the name under which to store the array
    :return: the numpy array
    """
    filepath = get_numpy_array_absolute_path(uuid, name)
    return numpy.load(filepath)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# Generated by Django 1.11.16 on 2018-12-21 10:56
# pylint: disable=invalid-name,too-few-public-methods
"""Migration for the update of the DbLog table. Addition of uuids"""

from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

from __future__ import print_function
import sys
import click

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from django.db import migrations, models
from aiida.backends.djsite.db.migrations import upgrade_schema_version
from aiida import settings
import aiida.common.utils

REVISION = '1.0.24'
DOWN_REVISION = '1.0.23'

# The values that will be exported for the log records that will be deleted
values_to_export = ['id', 'time', 'loggername', 'levelname', 'objpk', 'objname', 'message', 'metadata']

node_prefix = 'node.'
leg_workflow_prefix = 'aiida.workflows.user.'


def get_legacy_workflow_log_number(dblog_model):
    """ Get the number of the log records that correspond to legacy workflows """
    return dblog_model.objects.filter(objname__startswith=leg_workflow_prefix).count()


def get_unknown_entity_log_number(dblog_model):
    """ Get the number of the log records that correspond to unknown entities """
    return dblog_model.objects.exclude(objname__startswith=node_prefix).exclude(
        objname__startswith=leg_workflow_prefix).count()


def get_logs_with_no_nodes_number(dblog_model, dbnode_model):
    """ Get the number of the log records that don't correspond to a node """
    return dblog_model.objects.exclude(objpk__in=dbnode_model.objects.values('id')).filter(
        objname__startswith=node_prefix).values(*values_to_export).count()


def get_serialized_legacy_workflow_logs(dblog_model):
    """ Get the serialized log records that correspond to legacy workflows """
    from aiida.backends.sqlalchemy.utils import dumps_json
    queryset = dblog_model.objects.filter(objname__startswith=leg_workflow_prefix).values(*values_to_export)
    return dumps_json(list(queryset))


def get_serialized_unknown_entity_logs(dblog_model):
    """ Get the serialized log records that correspond to unknown entities """
    from aiida.backends.sqlalchemy.utils import dumps_json
    queryset = dblog_model.objects.exclude(objname__startswith=node_prefix).exclude(
        objname__startswith=leg_workflow_prefix).values(*values_to_export)
    return dumps_json(list(queryset))


def get_serialized_logs_with_no_nodes(dblog_model, dbnode_model):
    """ Get the serialized log records that don't correspond to a node """
    from aiida.backends.sqlalchemy.utils import dumps_json
    queryset = dblog_model.objects.exclude(objpk__in=dbnode_model.objects.values('id')).filter(
        objname__startswith=node_prefix).values(*values_to_export)
    return dumps_json(list(queryset))


def set_new_uuid(apps, _):
    """
    Set new UUIDs for all logs
    """
    DbLog = apps.get_model('db', 'DbLog')
    for log in DbLog.objects.all():
        log.uuid = aiida.common.utils.get_new_uuid()
        log.save(update_fields=['uuid'])


def export_and_clean_workflow_logs(apps, _):
    """
    Export the logs records that correspond to legacy workflows and to unknown entities.
    """
    from tempfile import NamedTemporaryFile

    DbLog = apps.get_model('db', 'DbLog')
    DbNode = apps.get_model('db', 'DbNode')

    lwf_number = get_legacy_workflow_log_number(DbLog)
    other_number = get_unknown_entity_log_number(DbLog)
    log_no_node_number = get_logs_with_no_nodes_number(DbLog, DbNode)

    # If there are no legacy workflow log records or log records of an unknown entity
    if lwf_number == 0 and other_number == 0 and log_no_node_number == 0:
        return

    if not settings.TESTING_MODE:
        click.echo('We found {} log records that correspond to legacy workflows and {} log records to correspond '
                   'to an unknown entity.'.format(lwf_number, other_number))
        click.echo(
            'These records will be removed from the database and exported to JSON files to the current directory).')
        proceed = click.confirm('Would you like to proceed?', default=True)
        if not proceed:
            sys.exit(1)

    delete_on_close = False
    if settings.TESTING_MODE:
        delete_on_close = True

    # Exporting the legacy workflow log records
    if lwf_number != 0:
        # Get the records and write them to file
        with NamedTemporaryFile(
                prefix='legagy_wf_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+') as handle:
            filename = handle.name
            handle.write(get_serialized_legacy_workflow_logs(DbLog))

        # If delete_on_close is False, we are running for the user and add additional message of file location
        if not delete_on_close:
            click.echo('Exported legacy workflow logs to {}'.format(filename))

        # Now delete the records
        DbLog.objects.filter(objname__startswith=leg_workflow_prefix).delete()

    # Exporting unknown log records
    if other_number != 0:
        # Get the records and write them to file
        with NamedTemporaryFile(
                prefix='unknown_entity_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+') as handle:
            filename = handle.name
            handle.write(get_serialized_unknown_entity_logs(DbLog))

        # If delete_on_close is False, we are running for the user and add additional message of file location
        if not delete_on_close:
            click.echo('Exported unexpected entity logs to {}'.format(filename))

        # Now delete the records
        DbLog.objects.exclude(objname__startswith=node_prefix).exclude(objname__startswith=leg_workflow_prefix).delete()

    # Exporting log records that don't correspond to nodes
    if log_no_node_number != 0:
        # Get the records and write them to file
        with NamedTemporaryFile(
                prefix='no_node_entity_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+') as handle:
            filename = handle.name
            handle.write(get_serialized_logs_with_no_nodes(DbLog, DbNode))

        # If delete_on_close is False, we are running for the user and add additional message of file location
        if not delete_on_close:
            click.echo('Exported entity logs that don\'t correspond to nodes to {}'.format(filename))

        # Now delete the records
        DbLog.objects.exclude(objpk__in=DbNode.objects.values('id')).filter(objname__startswith=node_prefix).delete()


def clean_dblog_metadata(apps, _):
    """
    Remove objpk and objname from the DbLog table metadata.
    """
    import json

    DbLog = apps.get_model('db', 'DbLog')
    for log in DbLog.objects.all():
        met = json.loads(log.metadata)
        if 'objpk' in met:
            del met['objpk']
        if 'objname' in met:
            del met['objname']
        log.metadata = json.dumps(met)
        log.save(update_fields=['metadata'])


def enrich_dblog_metadata(apps, _):
    """
    Add objpk and objname to the DbLog table metadata.
    """
    import json

    DbLog = apps.get_model('db', 'DbLog')
    for log in DbLog.objects.all():
        met = json.loads(log.metadata)
        if 'objpk' not in met:
            met['objpk'] = log.objpk
        if 'objname' not in met:
            met['objname'] = log.objname
        log.metadata = json.dumps(met)
        log.save(update_fields=['metadata'])


class Migration(migrations.Migration):
    """
    This migration updates the DbLog schema and adds UUID for correct export of the DbLog entries.
    More specifically, it adds UUIDS, it exports to files the not needed log entries (that correspond
    to legacy workflows and unknown entities), it creates a foreign key to the dbnode table, it
    transfers there the objpk data to the new dbnode column (just altering the objpk column and making
    it a foreign key when containing data, raised problems) and in the end objpk and objname columns
    are removed.
    """

    dependencies = [
        ('db', '0023_calc_job_option_attribute_keys'),
    ]

    operations = [
        # Export of the logs of the old workflows to a JSON file, there is no re-import
        # for the reverse migrations
        migrations.RunPython(export_and_clean_workflow_logs, reverse_code=migrations.RunPython.noop),

        # Removing objname and objpk from the metadata. The reverse migration adds the
        # objname and objpk to the metadata
        migrations.RunPython(clean_dblog_metadata, reverse_code=enrich_dblog_metadata),

        # The forward migration will not do anything for the objname, the reverse
        # migration will populate it with correct values
        migrations.RunSQL(
            '',
            reverse_sql='UPDATE db_dblog SET objname=db_dbnode.type '
            'FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk'),

        # Removal of the column objname, the reverse migration will add it
        migrations.RemoveField(model_name='dblog', name='objname'),

        # Creation of a new column called dbnode which is a foreign key to the dbnode table
        # The reverse migration will remove this column
        migrations.AddField(
            model_name='dblog',
            name='dbnode',
            field=models.ForeignKey(
                on_delete=models.deletion.CASCADE, related_name='dblogs', to='db.DbNode', blank=True, null=True),
        ),

        # Transfer of the data from the objpk to the node field
        # The reverse migration will do the inverse transfer
        migrations.RunSQL('UPDATE db_dblog SET dbnode_id=objpk', reverse_sql='UPDATE db_dblog SET objpk=dbnode_id'),

        # Now that all the data have been migrated, make the column not nullable and not blank.
        # A log record should always correspond to a node record
        migrations.AlterField(
            model_name='dblog',
            name='dbnode',
            field=models.ForeignKey(on_delete=models.deletion.CASCADE, related_name='dblogs', to='db.DbNode'),
        ),

        # Since the new column is created correctly, drop the old objpk column
        # The reverse migration will add the field
        migrations.RemoveField(model_name='dblog', name='objpk'),

        # This is the correct pattern to generate unique fields, see
        # https://docs.djangoproject.com/en/1.11/howto/writing-migrations/#migrations-that-add-unique-fields
        # The reverse migration will remove it
        migrations.AddField(
            model_name='dblog',
            name='uuid',
            field=models.UUIDField(default=aiida.common.utils.get_new_uuid, null=True),
        ),

        # Add unique UUIDs to the UUID field. There is no need for a reverse migration for a field
        # tha will be deleted
        migrations.RunPython(set_new_uuid, reverse_code=migrations.RunPython.noop),

        # Changing the column to unique
        migrations.AlterField(
            model_name='dblog',
            name='uuid',
            field=models.UUIDField(default=aiida.common.utils.get_new_uuid, unique=True),
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This module contains tests for UpfData and UpfData related functions.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import errno
import tempfile
import shutil
import os

from aiida.backends.testbase import AiidaTestCase
from aiida.common.exceptions import ParsingError
from aiida.orm.nodes.data.upf import parse_upf


class TestUpfParser(AiidaTestCase):
    """Tests UPF version / element_name parser function."""

    def setUp(self):
        """Setup a temporary directory to store UPF files."""

        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Destroy the temporary directory created."""

        try:
            shutil.rmtree(self.temp_dir)
        except OSError as exception:
            if exception.errno == errno.ENOENT:
                pass
            elif exception.errno == errno.ENOTDIR:
                os.remove(self.temp_dir)
            else:
                raise IOError(exception)

    # Test for parsing a regular UPF file for UPF-file-version 1
    def test_upf_version_one(self):
        """Check if parsing for regular UPF file (version 1) succeeds."""

        upf_filename = "O.test_file_v1.UPF"
        # regular upf file version 1 header
        upf_contents = u"\n".join([
            "<PP_INFO>"
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "O                     Element",
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # try to parse version / element name from UPF file contents
        parsed_data = parse_upf(path_to_upf, check_filename=True)
        # check that parsed data matches the expected one
        self.assertEqual(parsed_data['version'], '1')
        self.assertEqual(parsed_data['element'], 'O')

    # Test for parsing a regular UPF file for UPF-file-version 2
    def test_upf_version_two(self):
        """Check if parsing for regular UPF file (version 2) succeeds."""

        upf_filename = "Al.test_file_v2.UPF"
        # regular upf file version 2 header
        upf_contents = u"\n".join([
            "<UPF version=\"2.0.1\">",
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "element=\"Al\"",
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # try to parse version / element name from UPF file contents
        parsed_data = parse_upf(path_to_upf, check_filename=True)
        # check that parsed data matches the expected one
        self.assertEqual(parsed_data['version'], '2.0.1')
        self.assertEqual(parsed_data['element'], 'Al')

    # test for issue #2228 failing since additional header line shifted
    # the actual UPF version to the second line
    def test_additional_header_line(self):
        """Check if parsing succeeds if additional header line is present."""

        upf_filename = "Pt.test_file.UPF"
        # minimal contents required for parsing including additional header
        # file
        upf_contents = u"\n".join([
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>",
            "<UPF version=\"2.0.1\">",
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "element=\"Pt\"",
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # try to parse version / element name from UPF file contents
        parsed_data = parse_upf(path_to_upf, check_filename=True)
        # check that parsed data matches the expected one
        self.assertEqual(parsed_data['version'], '2.0.1')
        self.assertEqual(parsed_data['element'], 'Pt')

    def test_check_filename(self):
        """Test built-in check for if file name matches element"""

        upf_filename = "Al.test_file.UPF"
        # upf file header contents
        upf_contents = u"\n".join([
            "<UPF version=\"2.0.1\">",
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "element=\"Pt\"",
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # Check if parser raises the desired ParsingError
        with self.assertRaises(ParsingError):
            _ = parse_upf(path_to_upf, check_filename=True)

    def test_missing_element_upf_v2(self):
        """Test parsers exception on missing element name in UPF v2."""

        upf_filename = "Ab.test_file_missing_element_v2.UPF"
        # upf file header contents
        upf_contents = u"\n".join([
            "<UPF version=\"2.0.1\">",
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "element should be here but is missing",
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # Check if parser raises the desired ParsingError
        with self.assertRaises(ParsingError):
            _ = parse_upf(path_to_upf, check_filename=True)

    def test_invalid_element_upf_v2(self):
        """Test parsers exception on invalid element name in UPF v2."""

        upf_filename = "Ab.test_file_invalid_element_v2.UPF"
        # upf file header contents
        upf_contents = u"\n".join([
            "<UPF version=\"2.0.1\">",
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "element=\"Ab\""
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # Check if parser raises the desired ParsingError
        with self.assertRaises(ParsingError):
            _ = parse_upf(path_to_upf, check_filename=True)

    def test_missing_element_upf_v1(self):
        """Test parsers exception on missing element name in UPF v1."""

        upf_filename = "O.test_file_missing_element_v1.UPF"
        # upf file header contents
        upf_contents = u"\n".join([
            "<PP_INFO>"
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "element should be here but is missing",
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # Check if parser raises the desired ParsingError
        with self.assertRaises(ParsingError):
            _ = parse_upf(path_to_upf, check_filename=True)

    def test_invalid_element_upf_v1(self):
        """Test parsers exception on invalid element name in UPF v1."""

        upf_filename = "Ab.test_file_invalid_element_v1.UPF"
        # upf file header contents
        upf_contents = u"\n".join([
            "<PP_INFO>"
            "Human readable section is completely irrelevant for parsing!",
            "<PP_HEADER",
            "contents before element tag",
            "Ab                     Element",
            "contents following element tag",
            ">",
        ])
        path_to_upf = os.path.join(self.temp_dir, upf_filename)
        with open(path_to_upf, 'w') as upf_file:
            upf_file.write(upf_contents)
        # Check if parser raises the desired ParsingError
        with self.assertRaises(ParsingError):
            _ = parse_upf(path_to_upf, check_filename=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Module to get the backend instance from the Models instance
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

try:  # Python3
    from functools import singledispatch
except ImportError:  # Python2
    from singledispatch import singledispatch

# pylint: disable=cyclic-import

from aiida.backends.djsite.db import models
from aiida.orm.implementation.django import dummy_model as dummy_models

__all__ = ('get_backend_entity',)


@singledispatch
def get_backend_entity(dbmodel, backend):  # pylint: disable=unused-argument
    """
    Default get_backend_entity from DbModel

    :param dbmodel: the db model instance
    """
    raise TypeError('No corresponding AiiDA backend class exists for the DbModel instance {}'.format(
        dbmodel.__class__.__name__))


@get_backend_entity.register(models.DbUser)
def _(dbmodel, backend):
    """
    get_backend_entity for Django DbUser
    """
    from . import users
    return users.DjangoUser.from_dbmodel(dbmodel, backend)


@get_backend_entity.register(models.DbGroup)
def _(dbmodel, backend):
    """
    get_backend_entity for Django DbGroup
    """
    from . import groups
    return groups.DjangoGroup.from_dbmodel(dbmodel, backend)


@get_backend_entity.register(models.DbComputer)
def _(dbmodel, backend):
    """
    get_backend_entity for Django DbGroup
    """
    from . import computers
    return computers.DjangoComputer.from_dbmodel(dbmodel, backend)


@get_backend_entity.register(models.DbNode)
def _(dbmodel, backend):
    """
    get_backend_entity for Django DbNode. It will return an ORM instance since
    there is not Node backend entity yet.
    """
    from . import nodes
    return nodes.DjangoNode.from_dbmodel(dbmodel, backend)


@get_backend_entity.register(models.DbAuthInfo)
def _(dbmodel, backend):
    """
    get_backend_entity for Django DbAuthInfo
    """
    from . import authinfos
    return authinfos.DjangoAuthInfo.from_dbmodel(dbmodel, backend)


@get_backend_entity.register(models.DbComment)
def _(dbmodel, backend):
    from . import comments
    return comments.DjangoComment.from_dbmodel(dbmodel, backend)


@get_backend_entity.register(models.DbLog)
def _(dbmodel, backend):
    from . import logs
    return logs.DjangoLog.from_dbmodel(dbmodel, backend)


@get_backend_entity.register(dummy_models.DbUser)
def _(dbmodel, backend):
    """
    get_backend_entity for DummyModel DbUser.
    DummyModel instances are created when QueryBuilder queries the Django backend.
    """
    from . import users
    djuser_instance = models.DbUser(
        id=dbmodel.id,
        email=dbmodel.email,
        password=dbmodel.password,
        first_name=dbmodel.first_name,
        last_name=dbmodel.last_name,
        institution=dbmodel.institution,
        is_staff=dbmodel.is_staff,
        is_active=dbmodel.is_active,
        last_login=dbmodel.last_login,
        date_joined=dbmodel.date_joined)
    return users.DjangoUser.from_dbmodel(djuser_instance, backend)


@get_backend_entity.register(dummy_models.DbGroup)
def _(dbmodel, backend):
    """
    get_backend_entity for DummyModel DbGroup.
    DummyModel instances are created when QueryBuilder queries the Django backend.
    """
    from . import groups
    djgroup_instance = models.DbGroup(
        id=dbmodel.id,
        type_string=dbmodel.type_string,
        uuid=dbmodel.uuid,
        label=dbmodel.label,
        time=dbmodel.time,
        description=dbmodel.description,
        user_id=dbmodel.user_id,
    )
    return groups.DjangoGroup.from_dbmodel(djgroup_instance, backend)


@get_backend_entity.register(dummy_models.DbComputer)
def _(dbmodel, backend):
    """
    get_backend_entity for DummyModel DbComputer.
    DummyModel instances are created when QueryBuilder queries the Django backend.
    """
    from . import computers
    djcomputer_instance = models.DbComputer(
        id=dbmodel.id,
        uuid=dbmodel.uuid,
        name=dbmodel.name,
        hostname=dbmodel.hostname,
        description=dbmodel.description,
        enabled=dbmodel.enabled,
        transport_type=dbmodel.transport_type,
        scheduler_type=dbmodel.scheduler_type,
        transport_params=dbmodel.transport_params,
        metadata=dbmodel._metadata)  # pylint: disable=protected-access
    return computers.DjangoComputer.from_dbmodel(djcomputer_instance, backend)


@get_backend_entity.register(dummy_models.DbNode)
def _(dbmodel, backend):
    """
    get_backend_entity for DummyModel DbNode.
    DummyModel instances are created when QueryBuilder queries the Django backend.
    """
    djnode_instance = models.DbNode(
        id=dbmodel.id,
        node_type=dbmodel.node_type,
        process_type=dbmodel.process_type,
        uuid=dbmodel.uuid,
        ctime=dbmodel.ctime,
        mtime=dbmodel.mtime,
        label=dbmodel.label,
        description=dbmodel.description,
        dbcomputer_id=dbmodel.dbcomputer_id,
        user_id=dbmodel.user_id,
        public=dbmodel.public,
        nodeversion=dbmodel.nodeversion)

    from . import nodes
    return nodes.DjangoNode.from_dbmodel(djnode_instance, backend)


@get_backend_entity.register(dummy_models.DbAuthInfo)
def _(dbmodel, backend):
    """
    get_backend_entity for DummyModel DbAuthInfo.
    DummyModel instances are created when QueryBuilder queries the Django backend.
    """
    from . import authinfos
    djauthinfo_instance = models.DbAuthInfo(
        id=dbmodel.id,
        aiidauser_id=dbmodel.aiidauser_id,
        dbcomputer_id=dbmodel.dbcomputer_id,
        metadata=dbmodel._metadata,  # pylint: disable=protected-access
        auth_params=dbmodel.auth_params,
        enabled=dbmodel.enabled,
    )
    return authinfos.DjangoAuthInfo.from_dbmodel(djauthinfo_instance, backend)


@get_backend_entity.register(dummy_models.DbComment)
def _(dbmodel, backend):
    """
    Convert a dbcomment to the backend entity
    """
    from . import comments
    djcomment = models.DbComment(
        id=dbmodel.id,
        uuid=dbmodel.uuid,
        dbnode_id=dbmodel.dbnode_id,
        ctime=dbmodel.ctime,
        mtime=dbmodel.mtime,
        user_id=dbmodel.user_id,
        content=dbmodel.content)
    return comments.DjangoComment.from_dbmodel(djcomment, backend)


@get_backend_entity.register(dummy_models.DbLog)
def _(dbmodel, backend):
    """
    Convert a dbcomment to the backend entity
    """
    from . import logs
    djlog = models.DbLog(
        id=dbmodel.id,
        time=dbmodel.time,
        loggername=dbmodel.loggername,
        levelname=dbmodel.levelname,
        dbnode_id=dbmodel.dbnode_id,
        message=dbmodel.message,
        metadata=dbmodel._metadata  # pylint: disable=protected-access
    )
    return logs.DjangoLog.from_dbmodel(djlog, backend)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Setup script for aiida-core package."""

from __future__ import division
from __future__ import print_function

from __future__ import absolute_import
import sys
import json
from os import path
# pylint: disable=wrong-import-order
# Note: This speeds up command line scripts (e.g. verdi)
from utils import fastentrypoints  # pylint: disable=unused-import
from distutils.version import StrictVersion
from setuptools import setup, find_packages

if __name__ == '__main__':
    THIS_FOLDER = path.split(path.abspath(__file__))[0]

    # Ensure that pip is installed and the version is at least 10.0.0, which is required for the build process
    try:
        import pip
    except ImportError:
        print('Could not import pip, which is required for installation')
        sys.exit(1)

    PIP_REQUIRED_VERSION = '10.0.0'
    REQUIRED_VERSION = StrictVersion(PIP_REQUIRED_VERSION)
    INSTALLED_VERSION = StrictVersion(pip.__version__)

    if INSTALLED_VERSION < REQUIRED_VERSION:
        print('The installation requires pip>={}, whereas currently {} is installed'.format(
            REQUIRED_VERSION, INSTALLED_VERSION))
        sys.exit(1)

    with open(path.join(THIS_FOLDER, 'setup.json'), 'r') as info:
        SETUP_JSON = json.load(info)

    SETUP_JSON['extras_require']['testing'] \
        += SETUP_JSON['extras_require']['rest'] \
        + SETUP_JSON['extras_require']['atomic_tools']

    SETUP_JSON['extras_require']['docs'] \
        += SETUP_JSON['extras_require']['rest'] \
        + SETUP_JSON['extras_require']['atomic_tools']

    SETUP_JSON['extras_require']['all'] = list(
        {item for sublist in SETUP_JSON['extras_require'].values() for item in sublist if item != 'bpython'})

    setup(
        packages=find_packages(),
        long_description=open(path.join(THIS_FOLDER, 'README.md')).read(),
        long_description_content_type='text/markdown',
        **SETUP_JSON)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import sys
import io
import os
import shutil
import unittest
import tempfile
import numpy as np
import subprocess as sp

from contextlib import contextmanager
from six.moves import cStringIO as StringIO
from click.testing import CliRunner

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands.cmd_data import cmd_array
from aiida.cmdline.commands.cmd_data import cmd_bands
from aiida.cmdline.commands.cmd_data import cmd_cif
from aiida.cmdline.commands.cmd_data import cmd_parameter
from aiida.cmdline.commands.cmd_data import cmd_remote
from aiida.cmdline.commands.cmd_data import cmd_structure
from aiida.cmdline.commands.cmd_data import cmd_trajectory
from aiida.engine import calcfunction
from aiida.orm.nodes.data.cif import has_pycifrw
from aiida.orm import ArrayData, BandsData, CifData, KpointsData, Dict, RemoteData, StructureData, TrajectoryData


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestVerdiDataExportable:

    def __init__(self):
        pass

    NODE_ID_STR = "node_id"
    EMPTY_GROUP_ID_STR = 'empty_group_id'
    EMPTY_GROUP_NAME_STR = 'empty_group'
    NON_EMPTY_GROUP_ID_STR = 'non_empty_group_id'
    NON_EMPTY_GROUP_NAME_STR = 'non_empty_group'

    @unittest.skipUnless(has_pycifrw(), "Unable to import PyCifRW")
    def data_export_test(self, datatype, ids, supported_formats):
        """
        This method tests that the data listing works as expected with all
        possible flags and arguments for different datatypes.
        """

        from aiida.cmdline.commands.cmd_data.cmd_cif import cif_export
        from aiida.cmdline.commands.cmd_data.cmd_structure import structure_export
        from aiida.cmdline.commands.cmd_data.cmd_trajectory import trajectory_export

        datatype_mapping = {
            CifData: cif_export,
            StructureData: structure_export,
            TrajectoryData: trajectory_export,
        }

        if datatype is None or datatype not in datatype_mapping.keys():
            raise Exception("The listing of the objects {} is not supported".format(datatype))

        export_cmd = datatype_mapping[datatype]

        # Check that the simple command works as expected
        options = [str(ids[self.NODE_ID_STR])]
        res = self.cli_runner.invoke(export_cmd, options, catch_exceptions=False)
        self.assertEqual(res.exit_code, 0, "The command did not finish correctly")

        dump_flags = ['-F', '--format']
        for flag in dump_flags:
            for format in supported_formats:
                # with captured_output() as (out, err):
                options = [flag, format, str(ids[self.NODE_ID_STR])]
                res = self.cli_runner.invoke(export_cmd, options,
                                             catch_exceptions=False)
                self.assertEqual(res.exit_code, 0,
                                  "The command did not finish "
                                  "correctly. Output:\n{}".format(res.output))

        # Check that the output to file flags work correctly:
        # -o, --output
        output_flags = ['-o', '--output']
        for flag in output_flags:
            try:
                tmpd = tempfile.mkdtemp()
                filepath = os.path.join(tmpd, 'output_file.txt')
                options = [flag, filepath, str(ids[self.NODE_ID_STR])]
                res = self.cli_runner.invoke(export_cmd, options,
                                             catch_exceptions=False)
                self.assertEqual(res.exit_code, 0,
                                  "The command should finish correctly."
                                  "Output:\n{}".format(res.output))

                # Try to export it again. It should fail because the
                # file exists
                res = self.cli_runner.invoke(export_cmd, options, catch_exceptions=False)
                self.assertNotEquals(res.exit_code, 0, "The command should fail because the file already exists")

                # Now we force the export of the file and it should overwrite
                # existing files
                options = [flag, filepath, '-f', str(ids[self.NODE_ID_STR])]
                res = self.cli_runner.invoke(export_cmd, options,
                                             catch_exceptions=False)
                self.assertEqual(res.exit_code, 0,
                                  "The command should finish correctly."
                                  "Output: {}".format(res.output))
            finally:
                shutil.rmtree(tmpd)


class TestVerdiDataListable:

    def __init__(self):
        pass

    NODE_ID_STR = "node_id"
    EMPTY_GROUP_ID_STR = 'empty_group_id'
    EMPTY_GROUP_NAME_STR = 'empty_group'
    NON_EMPTY_GROUP_ID_STR = 'non_empty_group_id'
    NON_EMPTY_GROUP_NAME_STR = 'non_empty_group'

    def data_listing_test(self, datatype, search_string, ids):
        """
        This method tests that the data listing works as expected with all
        possible flags and arguments for different datatypes.
        """

        from aiida.cmdline.commands.cmd_data.cmd_cif import cif_list
        from aiida.cmdline.commands.cmd_data.cmd_structure import structure_list
        from aiida.cmdline.commands.cmd_data.cmd_trajectory import trajectory_list
        from aiida.cmdline.commands.cmd_data.cmd_bands import bands_list

        from aiida.cmdline.commands.cmd_data.cmd_structure import LIST_PROJECT_HEADERS as p_str
        from aiida.cmdline.commands.cmd_data.cmd_cif import LIST_PROJECT_HEADERS as p_cif
        from aiida.cmdline.commands.cmd_data.cmd_trajectory import LIST_PROJECT_HEADERS as p_tr
        from aiida.cmdline.commands.cmd_data.cmd_bands import LIST_PROJECT_HEADERS as p_bands

        headers_mapping = {CifData: p_cif, StructureData: p_str, TrajectoryData: p_tr, BandsData: p_bands}

        datatype_mapping = {
            CifData: cif_list,
            StructureData: structure_list,
            TrajectoryData: trajectory_list,
            BandsData: bands_list
        }

        if datatype is None or datatype not in datatype_mapping.keys():
            raise Exception("The listing of the objects {} is not supported".format(datatype))

        listing_cmd = datatype_mapping[datatype]
        project_headers = headers_mapping[datatype]

        # the output we are comparing to is un-decoded bytes, convert search strings:
        search_string_bytes = search_string.encode('utf-8')

        # Check that the normal listing works as expected
        res = self.cli_runner.invoke(listing_cmd, [], catch_exceptions=False)
        self.assertIn(search_string_bytes, res.stdout_bytes, 'The string {} was not found in the listing'
                      .format(search_string))

        # Check that the past days filter works as expected
        past_days_flags = ['-p', '--past-days']
        # past_days_flags = ['-p']
        for flag in past_days_flags:
            options = [flag, '1']
            res = self.cli_runner.invoke(listing_cmd, options, catch_exceptions=False)
            self.assertIn(search_string_bytes, res.stdout_bytes, 'The string {} was not found in the listing'
                          .format(search_string))

            options = [flag, '0']
            res = self.cli_runner.invoke(listing_cmd, options, catch_exceptions=False)
            self.assertNotIn(search_string_bytes, res.stdout_bytes, 'A not expected string {} was found in the listing'
                             .format(search_string))

        # Check that the group filter works as expected
        group_flags = ['-G', '--groups']
        for flag in group_flags:
            # Non empty group
            for non_empty in [self.NON_EMPTY_GROUP_NAME_STR, str(ids[self.NON_EMPTY_GROUP_ID_STR])]:
                options = [flag, non_empty]
                res = self.cli_runner.invoke(listing_cmd, options, catch_exceptions=False)
                self.assertIn(search_string_bytes, res.stdout_bytes, 'The string {} was not found in the listing')

            # Empty group
            for empty in [self.EMPTY_GROUP_NAME_STR, str(ids[self.EMPTY_GROUP_ID_STR])]:
                options = [flag, empty]
                res = self.cli_runner.invoke(listing_cmd, options, catch_exceptions=False)
                self.assertNotIn(search_string_bytes, res.stdout_bytes,
                                 'A not expected string {} was found in the listing')

            # Group combination
            for non_empty in [self.NON_EMPTY_GROUP_NAME_STR, str(ids[self.NON_EMPTY_GROUP_ID_STR])]:
                for empty in [self.EMPTY_GROUP_NAME_STR, str(ids[self.EMPTY_GROUP_ID_STR])]:
                    options = [flag, non_empty, empty]
                    res = self.cli_runner.invoke(listing_cmd, options, catch_exceptions=False)
                    self.assertIn(search_string_bytes, res.stdout_bytes, 'The string {} was not found in the listing')

        # Check raw flag
        raw_flags = ['-r', '--raw']
        for flag in raw_flags:
            options = [flag]
            res = self.cli_runner.invoke(listing_cmd, options, catch_exceptions=False)
            for header in project_headers:
                self.assertNotIn(header.encode('utf-8'), res.stdout_bytes)


class TestVerdiData(AiidaTestCase):
    """
    Testing reachability of the verdi data subcommands
    """

    @classmethod
    def setUpClass(cls):
        super(TestVerdiData, cls).setUpClass()

    def setUp(self):
        pass

    def test_reachable(self):
        """
        Testing reachability of the following commands:
        verdi data array
        verdi data bands
        verdi data cif
        verdi data parameter
        verdi data remote
        verdi data structure
        verdi data trajectory
        verdi data upf
        """
        subcommands = ['array', 'bands', 'cif', 'parameter', 'remote', 'structure', 'trajectory', 'upf']
        for sub_cmd in subcommands:
            output = sp.check_output(['verdi', 'data', sub_cmd, '--help'])
            self.assertIn(b'Usage:', output, "Sub-command verdi data {} --help failed.".format(sub_cmd))


class TestVerdiDataArray(AiidaTestCase):
    """
    Testing verdi data array
    """

    @classmethod
    def setUpClass(cls):
        super(TestVerdiDataArray, cls).setUpClass()

    def setUp(self):
        self.a = ArrayData()
        self.a.set_array('test_array', np.array([0, 1, 3]))
        self.a.store()

        self.cli_runner = CliRunner()

    def test_arrayshowhelp(self):
        output = sp.check_output(['verdi', 'data', 'array', 'show', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data array show --help failed.")

    def test_arrayshow(self):
        # with captured_output() as (out, err):
        options = [str(self.a.id)]
        res = self.cli_runner.invoke(cmd_array.array_show, options, catch_exceptions=False)
        self.assertEqual(res.exit_code, 0, "The command did not finish correctly")


class TestVerdiDataBands(AiidaTestCase, TestVerdiDataListable):
    """
    Testing verdi data bands
    """

    @staticmethod
    def create_structure_bands():
        alat = 4.  # angstrom
        cell = [
            [
                alat,
                0.,
                0.,
            ],
            [
                0.,
                alat,
                0.,
            ],
            [
                0.,
                0.,
                alat,
            ],
        ]
        s = StructureData(cell=cell)
        s.append_atom(position=(0., 0., 0.), symbols='Fe')
        s.append_atom(position=(alat / 2., alat / 2., alat / 2.), symbols='O')
        s.store()

        @calcfunction
        def connect_structure_bands(structure):
            alat = 4.
            cell = np.array([
                [alat, 0., 0.],
                [0., alat, 0.],
                [0., 0., alat],
            ])

            k = KpointsData()
            k.set_cell(cell)
            k.set_kpoints_path([('G', 'M', 2)])

            b = BandsData()
            b.set_kpointsdata(k)
            b.set_bands([[1.0, 2.0], [3.0, 4.0]])

            return b

        b = connect_structure_bands(s)

        # Create 2 groups and add the data to one of them
        g_ne = orm.Group(label='non_empty_group')
        g_ne.store()
        g_ne.add_nodes(b)

        g_e = orm.Group(label='empty_group')
        g_e.store()

        return {
            TestVerdiDataListable.NODE_ID_STR: b.id,
            TestVerdiDataListable.NON_EMPTY_GROUP_ID_STR: g_ne.id,
            TestVerdiDataListable.EMPTY_GROUP_ID_STR: g_e.id
        }

    @classmethod
    def setUpClass(cls):
        super(TestVerdiDataBands, cls).setUpClass()
        cls.ids = cls.create_structure_bands()

    def setUp(self):
        self.cli_runner = CliRunner()

    def test_bandsshowhelp(self):
        output = sp.check_output(['verdi', 'data', 'bands', 'show', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data bands show --help failed.")

    def test_bandlistshelp(self):
        output = sp.check_output(['verdi', 'data', 'bands', 'list', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data bands show --help failed.")

    def test_bandslist(self):
        self.data_listing_test(BandsData, 'FeO', self.ids)

    def test_bandexporthelp(self):
        output = sp.check_output(['verdi', 'data', 'bands', 'export', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data bands export --help failed.")

    def test_bandsexport(self):
        options = [str(self.ids[TestVerdiDataListable.NODE_ID_STR])]
        res = self.cli_runner.invoke(cmd_bands.bands_export, options, catch_exceptions=False)
        self.assertEqual(res.exit_code, 0, 'The command did not finish correctly')
        self.assertIn(b"[1.0, 3.0]", res.stdout_bytes, 'The string [1.0, 3.0] was not found in the bands' 'export')


class TestVerdiDataParameter(AiidaTestCase):
    """
    Testing verdi data parameter
    """

    @classmethod
    def setUpClass(cls):
        super(TestVerdiDataParameter, cls).setUpClass()

    def setUp(self):
        self.p = Dict()
        self.p.set_dict({'a': 1, 'b': 2})
        self.p.store()

        self.cli_runner = CliRunner()

    def test_parametershowhelp(self):
        output = sp.check_output(['verdi', 'data', 'parameter', 'show', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data parameter show --help failed.")

    def test_parametershow(self):
        supported_formats = ['json_date']
        for format in supported_formats:
            options = [str(self.p.id)]
            res = self.cli_runner.invoke(cmd_parameter.parameter_show, options, catch_exceptions=False)
            self.assertEqual(res.exit_code, 0, "The command verdi data parameter show did not" " finish correctly")
        self.assertIn(b'"a": 1', res.stdout_bytes, 'The string "a": 1 was not found in the output'
                                                   ' of verdi data parameter show')


class TestVerdiDataRemote(AiidaTestCase):
    """
    Testing verdi data remote
    """

    @classmethod
    def setUpClass(cls):
        from aiida import orm
        super(TestVerdiDataRemote, cls).setUpClass()
        user = orm.User.objects.get_default()
        authinfo = orm.AuthInfo(cls.computer, user)
        authinfo.store()

    def setUp(self):
        comp = self.computer
        self.r = RemoteData()
        p = tempfile.mkdtemp()
        self.r.set_remote_path(p)
        with io.open(p + '/file.txt', 'w', encoding='utf8') as fhandle:
            fhandle.write(u"test string")
        self.r.computer = comp
        self.r.store()

        self.cli_runner = CliRunner()

    def test_remoteshowhelp(self):
        output = sp.check_output(['verdi', 'data', 'remote', 'show', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data remote show --help failed.")

    def test_remoteshow(self):
        options = [str(self.r.id)]
        res = self.cli_runner.invoke(cmd_remote.remote_show, options, catch_exceptions=False)
        self.assertEqual(res.exit_code, 0, "The command verdi data remote show did not" " finish correctly")
        self.assertIn(b'Remote computer name:', res.stdout_bytes,
                      'The string "Remote computer name:" was not found in the'
                      ' output of verdi data remote show')
        self.assertIn(b'Remote folder full path:', res.stdout_bytes,
                      'The string "Remote folder full path:" was not found in the'
                      ' output of verdi data remote show')

    def test_remotelshelp(self):
        output = sp.check_output(['verdi', 'data', 'remote', 'ls', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data remote ls --help failed.")

    def test_remotels(self):
        options = ['--long', str(self.r.id)]
        res = self.cli_runner.invoke(cmd_remote.remote_ls, options, catch_exceptions=False)
        self.assertEqual(res.exit_code, 0, "The command verdi data remote ls did not" " finish correctly")
        self.assertIn(b'file.txt', res.stdout_bytes, 'The file "file.txt" was not found in the output'
                                                     ' of verdi data remote ls')

    def test_remotecathelp(self):
        output = sp.check_output(['verdi', 'data', 'remote', 'cat', '--help'])
        self.assertIn(b'Usage:', output, "Sub-command verdi data remote cat --help failed.")

    def test_remotecat(self):
        options = [str(self.r.id), 'file.txt']
        res = self.cli_runner.invoke(cmd_remote.remote_cat, options, catch_exceptions=False)
        self.assertEqual(res.exit_code, 0, "The command verdi data parameter cat did not" " finish correctly")
        self.assertIn(b'test string', res.stdout_bytes, 'The string "test string" was not found in the output'
                                                        ' of verdi data remote cat file.txt')


class TestVerdiDataTrajectory(AiidaTestCase, TestVerdiDataListable, TestVerdiDataExportable):

    @staticmethod
    def create_trajectory_data():
        import numpy

        # Create a node with two arrays
        n = TrajectoryData()

        # I create sample data
        stepids = numpy.array([60, 70])
        times = stepids * 0.01
        cells = numpy.array([[[
            2.,
            0.,
            0.,
        ], [
            0.,
            2.,
            0.,
        ], [
            0.,
            0.,
            2.,
        ]], [[
            3.,
            0.,
            0.,
        ], [
            0.,
            3.,
            0.,
        ], [
            0.,
            0.,
            3.,
        ]]])
        symbols = ['H', 'O', 'C']
        positions = numpy.array([[[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]], [[0., 0., 0.], [0.5, 0.5, 0.5],
                                                                                    [1.5, 1.5, 1.5]]])
        velocities = numpy.array([[[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]], [[0.5, 0.5, 0.5], [0.5, 0.5, 0.5],
                                                                               [-0.5, -0.5, -0.5]]])

        # I set the node
        n.set_trajectory(
            stepids=stepids, cells=cells, symbols=symbols, positions=positions, times=times, velocities=velocities)

        n.store()

        # Create 2 groups and add the data to one of them
        g_ne = orm.Group(label='non_empty_group')
        g_ne.store()
        g_ne.add_nodes(n)

        g_e = orm.Group(label='empty_group')
        g_e.store()

        return {
            TestVerdiDataListable.NODE_ID_STR: n.id,
            TestVerdiDataListable.NON_EMPTY_GROUP_ID_STR: g_ne.id,
            TestVerdiDataListable.EMPTY_GROUP_ID_STR: g_e.id
        }

    @classmethod
    def setUpClass(cls):
        from aiida import orm
        super(TestVerdiDataTrajectory, cls).setUpClass()
        orm.Computer(
            name='comp',
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida').store()
        cls.ids = cls.create_trajectory_data()

    def setUp(self):
        self.comp = self.computer
        self.runner = CliRunner()
        self.this_folder = os.path.dirname(__file__)
        self.this_file = os.path.basename(__file__)

        self.cli_runner = CliRunner()

    def test_deposithelp(self):
        res = self.runner.invoke(cmd_trajectory.trajectory_deposit, ['--help'])
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data trajectory deposit --help')

    def test_showhelp(self):
        res = self.runner.invoke(cmd_trajectory.trajectory_show, ['--help'])
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data trajecotry show --help')

    def test_list(self):
        self.data_listing_test(TrajectoryData, str(self.ids[TestVerdiDataListable.NODE_ID_STR]), self.ids)

    @unittest.skipUnless(has_pycifrw(), "Unable to import PyCifRW")
    def test_export(self):
        from aiida.cmdline.commands.cmd_data.cmd_trajectory import EXPORT_FORMATS, trajectory_export

        new_supported_formats = list(EXPORT_FORMATS)
        # TCOD export needs special arguments
        new_supported_formats.remove('tcod')
        self.data_export_test(TrajectoryData, self.ids, new_supported_formats)
        # Check independently the TCOD export that needs special arguments
        dump_flags = ['-F', '--format']
        for flag in dump_flags:
            options = [flag, 'tcod', '-i', '0', str(self.ids[self.NODE_ID_STR])]
            res = self.cli_runner.invoke(trajectory_export, options, catch_exceptions=False)
            self.assertEqual(res.exit_code, 0, 'The command did not finish correctly')


class TestVerdiDataStructure(AiidaTestCase, TestVerdiDataListable, TestVerdiDataExportable):

    @staticmethod
    def create_structure_data():
        alat = 4.  # angstrom
        cell = [
            [
                alat,
                0.,
                0.,
            ],
            [
                0.,
                alat,
                0.,
            ],
            [
                0.,
                0.,
                alat,
            ],
        ]

        # BaTiO3 cubic structure
        struc = StructureData(cell=cell)
        struc.append_atom(position=(0., 0., 0.), symbols='Ba')
        struc.append_atom(position=(alat / 2., alat / 2., alat / 2.), symbols='Ti')
        struc.append_atom(position=(alat / 2., alat / 2., 0.), symbols='O')
        struc.append_atom(position=(alat / 2., 0., alat / 2.), symbols='O')
        struc.append_atom(position=(0., alat / 2., alat / 2.), symbols='O')
        struc.store()

        # Create 2 groups and add the data to one of them
        g_ne = orm.Group(label='non_empty_group')
        g_ne.store()
        g_ne.add_nodes(struc)

        g_e = orm.Group(label='empty_group')
        g_e.store()

        return {
            TestVerdiDataListable.NODE_ID_STR: struc.id,
            TestVerdiDataListable.NON_EMPTY_GROUP_ID_STR: g_ne.id,
            TestVerdiDataListable.EMPTY_GROUP_ID_STR: g_e.id
        }

    @classmethod
    def setUpClass(cls):
        from aiida import orm

        super(TestVerdiDataStructure, cls).setUpClass()
        orm.Computer(
            name='comp',
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida').store()
        cls.ids = cls.create_structure_data()

    def setUp(self):
        self.comp = self.computer
        self.runner = CliRunner()
        self.this_folder = os.path.dirname(__file__)
        self.this_file = os.path.basename(__file__)

        self.cli_runner = CliRunner()

    def test_importhelp(self):
        res = self.runner.invoke(cmd_structure.structure_import, ['--help'])
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data import --help')

    def test_import(self):
        xyzcontent = '''
        2

        Fe     0.0 0.0 0.0
        O      2.0 2.0 2.0
        '''
        with tempfile.NamedTemporaryFile(mode='w+') as fhandle:
            fhandle.write(xyzcontent)
            fhandle.flush()
            options = [
                fhandle.name,
                '--format',
                'xyz',
                '--vacuum-factor',
                '1.0',
                '--vacuum-addition',
                '10.0',
                '--pbc',
                '1',
                '1',
                '1',
            ]
            res = self.cli_runner.invoke(cmd_structure.structure_import, options, catch_exceptions=False)
            self.assertIn(b'PK = None', res.stdout_bytes, 'The string "PK = None" was not found in the output'
                                                          ' of verdi data structure import with --store option.')
            options.append('--store')
            res = self.cli_runner.invoke(cmd_structure.structure_import, options, catch_exceptions=False)
            self.assertIn(b'Succesfully imported', res.stdout_bytes,
                          'The string "Succesfully imported" was not found in the output'
                          ' of verdi data structure import.')

    def test_showhelp(self):
        res = self.runner.invoke(cmd_structure.structure_import, ['--help'])
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data show --help')

    def test_deposithelp(self):
        res = self.runner.invoke(cmd_structure.structure_import, ['--help'])
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data show --help')

    def test_list(self):
        self.data_listing_test(StructureData, 'BaO3Ti', self.ids)

    def test_export(self):
        from aiida.cmdline.commands.cmd_data.cmd_structure import EXPORT_FORMATS
        self.data_export_test(StructureData, self.ids, EXPORT_FORMATS)


@unittest.skipUnless(has_pycifrw(), "Unable to import PyCifRW")
class TestVerdiDataCif(AiidaTestCase, TestVerdiDataListable, TestVerdiDataExportable):
    valid_sample_cif_str = u'''
        data_test
        _cell_length_a    10
        _cell_length_b    10
        _cell_length_c    10
        _cell_angle_alpha 90
        _cell_angle_beta  90
        _cell_angle_gamma 90
        _chemical_formula_sum 'C O2'
        loop_
        _atom_site_label
        _atom_site_fract_x
        _atom_site_fract_y
        _atom_site_fract_z
        _atom_site_attached_hydrogens
        C 0 0 0 0
        O 0.5 0.5 0.5 .
        H 0.75 0.75 0.75 0
    '''

    @classmethod
    def create_cif_data(cls):
        with tempfile.NamedTemporaryFile(mode='w+') as fhandle:
            filename = fhandle.name
            fhandle.write(cls.valid_sample_cif_str)
            fhandle.flush()
            a = CifData(filepath=filename, source={'version': '1234', 'db_name': 'COD', 'id': '0000001'})
            a.store()

            g_ne = orm.Group(label='non_empty_group')
            g_ne.store()
            g_ne.add_nodes(a)

            g_e = orm.Group(label='empty_group')
            g_e.store()

        return {
            TestVerdiDataListable.NODE_ID_STR: a.id,
            TestVerdiDataListable.NON_EMPTY_GROUP_ID_STR: g_ne.id,
            TestVerdiDataListable.EMPTY_GROUP_ID_STR: g_e.id
        }

    @classmethod
    def setUpClass(cls):
        from aiida import orm
        super(TestVerdiDataCif, cls).setUpClass()
        orm.Computer(
            name='comp',
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida').store()
        cls.ids = cls.create_cif_data()

    def setUp(self):
        super(TestVerdiDataCif, self).setUp()
        self.comp = self.computer
        self.runner = CliRunner()
        self.this_folder = os.path.dirname(__file__)
        self.this_file = os.path.basename(__file__)

        self.cli_runner = CliRunner()

    def test_list(self):
        """
        This method tests that the Cif listing works as expected with all
        possible flags and arguments.
        """
        self.data_listing_test(CifData, 'C O2', self.ids)

    def test_showhelp(self):
        options = ['--help']
        res = self.cli_runner.invoke(cmd_cif.cif_show, options, catch_exceptions=False)
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data show help')

    def test_deposithelp(self):
        options = ['--help']
        res = self.cli_runner.invoke(cmd_cif.cif_deposit, options, catch_exceptions=False)
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data show deposit')

    def test_importhelp(self):
        options = ['--help']
        res = self.cli_runner.invoke(cmd_cif.cif_import, options, catch_exceptions=False)
        self.assertIn(b'Usage:', res.stdout_bytes, 'The string "Usage: " was not found in the output'
                                                   ' of verdi data import help')

    def test_import(self):
        with tempfile.NamedTemporaryFile(mode='w+') as fhandle:
            fhandle.write(self.valid_sample_cif_str)
            fhandle.flush()
            options = [fhandle.name]
            res = self.cli_runner.invoke(cmd_cif.cif_import, options, catch_exceptions=False)
            self.assertIn(b'imported uuid', res.stdout_bytes, 'The string "imported uuid" was not found in the output'
                                                              ' of verdi data import.')

    def test_export(self):
        """
        This method checks if the Cif export works as expected with all
        possible flags and arguments.
        """
        from aiida.cmdline.commands.cmd_data.cmd_cif import EXPORT_FORMATS
        self.data_export_test(CifData, self.ids, EXPORT_FORMATS)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This module provides deposit functionality to all data types
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click
from aiida.cmdline.params import options, types

DEPOSIT_OPTIONS = [
    options.CODE(help='Code to be used for the deposition.'),
    options.COMPUTER(help='Computer to be used for deposition.'),
    click.option(
        '-d', '--database', type=click.Choice(['tcod']), default='tcod', help='Label of the database for deposition.'),
    click.option(
        '--deposition-type',
        type=click.Choice(['published', 'prepublication', 'personal']),
        default='published',
        help='Type of the deposition.'),
    click.option('-u', '--username', type=click.STRING, default=None, help="Depositor's username."),
    click.option('-p', '--password', is_flag=True, default=False, help="Prompt for depositor's password."),
    click.option('--user-email', type=click.STRING, default=None, help="Depositor's e-mail address."),
    click.option('--title', type=click.STRING, default=None, help="Title of the publication."),
    click.option('--author-name', type=click.STRING, default=None, help="Full name of the publication author."),
    click.option('--author-email', type=click.STRING, default=None, help="E-mail address of the publication author."),
    click.option('--url', type=click.STRING, default=None, help="URL of the deposition API."),
    click.option(
        '--parameters',
        type=types.DataParamType(sub_classes=('aiida.data:dict',)),
        help="Dict to be exported alongside the to be deposited Data node. By default, if "
        "the node originates from a calculation with single Dict in the output, aforementioned "
        "Dict is picked automatically. Instead, the option is used in the case the calculation produces "
        "more than a single instance of Dict."),
    click.option(
        '--replace', type=click.INT, default=None, help="ID of the structure to be redeposited (replaced), if any."),
    click.option(
        '-m',
        '--message',
        type=click.STRING,
        default=None,
        help="Description of the change (relevant for redepositions only)."),
    click.option(
        '--reduce-symmetry/--no-reduce-symmetry',
        'reduce_symmetry',
        is_flag=True,
        default=None,
        help='Do (default) or do not perform symmetry reduction.'),
    click.option(
        '--dump-aiida-database/--no-dump-aiida-database',
        'dump_aiida_database',
        is_flag=True,
        default=None,
        help='Export (default) or do not export AiiDA database to the CIF file.'),
    click.option(
        '--exclude-external-contents/--no-exclude-external-contents',
        'exclude_external_contents',
        is_flag=True,
        default=None,
        help='Do not (default) or do save the contents for external resources even if URIs are provided'),
    click.option(
        '--gzip/--no-gzip', 'gzip', is_flag=True, default=None, help='Do or do not (default) gzip large files.'),
    click.option(
        '--gzip-threshold',
        type=click.INT,
        default=None,
        help="Specify the minimum size of exported file which should be gzipped."),
]


def deposit_options(func):
    for option in reversed(DEPOSIT_OPTIONS):
        func = option(func)

    return func


def data_deposit_tcod(node, deposit_type, parameters=None, **kwargs):
    """
    Deposition plugin for TCOD.
    """
    from aiida.tools.dbexporters.tcod import deposit
    return deposit(node, deposit_type, parameters=parameters, **kwargs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""DbGroup class: Rename name with label and type with type_string

Revision ID: b8b23ddefad4
Revises: 239cea6d2452
Create Date: 2018-12-06 15:25:32.865136

"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
# pylint: disable=no-name-in-module,import-error
from alembic import op

# revision identifiers, used by Alembic.
revision = 'b8b23ddefad4'
down_revision = '239cea6d2452'
branch_labels = None
depends_on = None


def upgrade():
    """The upgrade migration actions."""
    # dropping
    op.drop_constraint('db_dbgroup_name_type_key', 'db_dbgroup')
    op.drop_index('ix_db_dbgroup_name', 'db_dbgroup')
    op.drop_index('ix_db_dbgroup_type', 'db_dbgroup')

    # renaming
    op.alter_column('db_dbgroup', 'name', new_column_name='label')
    op.alter_column('db_dbgroup', 'type', new_column_name='type_string')

    # creating
    op.create_unique_constraint('db_dbgroup_label_type_string_key', 'db_dbgroup', ['label', 'type_string'])
    op.create_index('ix_db_dbgroup_label', 'db_dbgroup', ['label'])
    op.create_index('ix_db_dbgroup_type_string', 'db_dbgroup', ['type_string'])


def downgrade():
    """The downgrade migration actions."""
    # dropping
    op.drop_constraint('db_dbgroup_label_type_string_key', 'db_dbgroup')
    op.drop_index('ix_db_dbgroup_label', 'db_dbgroup')
    op.drop_index('ix_db_dbgroup_type_string', 'db_dbgroup')

    # renaming
    op.alter_column('db_dbgroup', 'label', new_column_name='name')
    op.alter_column('db_dbgroup', 'type_string', new_column_name='type')

    # creating
    op.create_unique_constraint('db_dbgroup_name_type_key', 'db_dbgroup', ['name', 'type'])
    op.create_index('ix_db_dbgroup_name', 'db_dbgroup', ['name'])
    op.create_index('ix_db_dbgroup_type', 'db_dbgroup', ['type'])



#!/usr/bin/env runaiida
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This example implements exactly the same functionality as seen in the basic WorkChain example, except in this case it
utilizes calcfunctions instead of workchains.
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from aiida.engine import calcfunction
from aiida.orm import Float, Int


@calcfunction
def sum(a, b):
    return a + b


@calcfunction
def product(a, b):
    return a * b


@calcfunction
def sumproduct(a, b, c):
    return product(sum(a, b), c)


def main():
    a = Float(3.14)
    b = Int(4)
    c = Int(6)

    results = sum(a, b)
    print('Result of sum: {}'.format(results))

    results = product(a, b)
    print('Result of product: {}'.format(results))

    results = sumproduct(a, b, c)
    print('Result of sumproduct: {}'.format(results))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-lines
"""Tests for the migration engine (Alembic) as well as for the AiiDA migrations for SQLAlchemy."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from contextlib import contextmanager

import os
from alembic import command
from alembic.config import Config
from six.moves import zip

from aiida.backends import sqlalchemy as sa
from aiida.backends.general.migrations import utils
from aiida.backends.sqlalchemy import utils as sqlalchemy_utils
from aiida.backends.sqlalchemy.models.base import Base
from aiida.backends.sqlalchemy.tests.test_utils import new_database
from aiida.backends.sqlalchemy.utils import flag_modified
from aiida.backends.testbase import AiidaTestCase


class TestMigrationsSQLA(AiidaTestCase):
    """
    This class contains tests for the migration mechanism of SQLAlchemy called
    alembic. It checks if the migrations can be applied and removed correctly.
    """
    # The path to the folder that contains the migration configuration (the
    # actual configuration - not the testing)
    migr_method_dir_path = None
    # The path of the migration configuration (the actual configuration - not
    # the testing)
    alembic_dpath = None

    migrate_from = None
    migrate_to = None

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Prepare the test class with the alembivc configuration
        """
        super(TestMigrationsSQLA, cls).setUpClass(*args, **kwargs)
        cls.migr_method_dir_path = os.path.dirname(os.path.realpath(sqlalchemy_utils.__file__))
        alembic_dpath = os.path.join(cls.migr_method_dir_path, sqlalchemy_utils.ALEMBIC_REL_PATH)
        cls.alembic_cfg = Config()
        cls.alembic_cfg.set_main_option('script_location', alembic_dpath)

    def setUp(self):
        """
        Go to the migrate_from revision, apply setUpBeforeMigration, then
        run the migration.
        """
        super(TestMigrationsSQLA, self).setUp()
        from aiida.orm import autogroup

        self.current_autogroup = autogroup.current_autogroup
        autogroup.current_autogroup = None
        assert self.migrate_from and self.migrate_to, \
            "TestCase '{}' must define migrate_from and migrate_to properties".format(type(self).__name__)

        try:
            self.migrate_db_down(self.migrate_from)
            self.setUpBeforeMigration()
            self.migrate_db_up(self.migrate_to)
        except Exception:
            # Bring back the DB to the correct state if this setup part fails
            self._reset_database_and_schema()
            raise

    def migrate_db_up(self, destination):
        """
        Perform a migration upwards (upgrade) with alembic

        :param destination: the name of the destination migration
        """
        # Undo all previous real migration of the database
        with sa.engine.connect() as connection:
            self.alembic_cfg.attributes['connection'] = connection  # pylint: disable=unsupported-assignment-operation
            command.upgrade(self.alembic_cfg, destination)

    def migrate_db_down(self, destination):
        """
        Perform a migration downwards (downgrade) with alembic

        :param destination: the name of the destination migration
        """
        with sa.engine.connect() as connection:
            self.alembic_cfg.attributes['connection'] = connection  # pylint: disable=unsupported-assignment-operation
            command.downgrade(self.alembic_cfg, destination)

    def tearDown(self):
        """
        Resets both the database content and the schema to prepare for the
        next test
        """
        from aiida.orm import autogroup
        self._reset_database_and_schema()
        autogroup.current_autogroup = self.current_autogroup
        super(TestMigrationsSQLA, self).tearDown()

    def setUpBeforeMigration(self):  # pylint: disable=invalid-name
        """
        Anything to do before running the migrations.
        This is typically implemented in test subclasses.
        """

    def _reset_database_and_schema(self):
        """
        Bring back the DB to the correct state.

        It is important to also reset the database content to avoid hanging
        of tests.
        """
        self.reset_database()
        self.migrate_db_up("head")

    @property
    def current_rev(self):
        """
        Utility method to get the current revision string
        """
        from alembic.migration import MigrationContext  # pylint: disable=import-error
        with sa.engine.begin() as connection:
            context = MigrationContext.configure(connection)
            current_rev = context.get_current_revision()
        return current_rev

    @staticmethod
    def get_auto_base():
        """
        Return the automap_base class that automatically inspects the current
        database and return SQLAlchemy Models.

        Note that these are NOT the ones in AiiDA SQLAlchemy models, so do not
        have the special methods that we define there (like .save()).
        """
        from alembic.migration import MigrationContext  # pylint: disable=import-error
        from sqlalchemy.ext.automap import automap_base  # pylint: disable=import-error,no-name-in-module

        with sa.engine.begin() as connection:
            context = MigrationContext.configure(connection)
            bind = context.bind

            base = automap_base()
            # reflect the tables
            base.prepare(bind.engine, reflect=True)

            return base

    @staticmethod
    @contextmanager
    def get_session():
        """
        Return a session that is properly closed after use.
        """
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        with sa.engine.begin() as connection:
            session = Session(connection.engine)
            yield session
            session.close()

    def get_current_table(self, table_name):
        """
        Return a Model instantiated at the correct migration.
        Note that this is obtained by inspecting the database and not
        by looking into the models file. So, special methods possibly defined
        in the models files/classes are not present.

        For instance, you can do::

          DbGroup = self.get_current_table('db_dbgroup')

        :param table_name: the name of the table.
        """
        base = self.get_auto_base()
        return getattr(base.classes, table_name)

    @staticmethod
    def get_node_array(node, name):
        return utils.load_numpy_array_from_repository(node.uuid, name)

    @staticmethod
    def set_node_array(node, name, array):
        """Store a new numpy array inside a node. Possibly overwrite the array if it already existed.

        Internally, it stores a name.npy file in numpy format.

        :param name: The name of the array.
        :param array: The numpy array to store.
        """
        utils.store_numpy_array_in_repository(node.uuid, name, array)
        attributes = node.attributes
        if attributes is None:
            attributes = {}
        attributes['array|{}'.format(name)] = list(array.shape)
        node.attributes = attributes
        flag_modified(node, 'attributes')


class TestBackwardMigrationsSQLA(TestMigrationsSQLA):
    """
    This is the equivalent of TestMigrationsSQLA for backward migrations.
    It assumes that the migrate_from revision is higher in the hierarchy
    than the migrate_to revision.
    """

    def setUp(self):
        """
        Go to the migrate_from revision, apply setUpBeforeMigration, then
        run the migration.
        """
        AiidaTestCase.setUp(self)  # pylint: disable=bad-super-call
        from aiida.orm import autogroup

        self.current_autogroup = autogroup.current_autogroup
        autogroup.current_autogroup = None
        assert self.migrate_from and self.migrate_to, \
            "TestCase '{}' must define migrate_from and migrate_to properties".format(type(self).__name__)

        try:
            self.migrate_db_down(self.migrate_from)
            self.setUpBeforeMigration()
            self.migrate_db_down(self.migrate_to)
        except Exception:
            # Bring back the DB to the correct state if this setup part fails
            self._reset_database_and_schema()
            raise


class TestMigrationEngine(TestMigrationsSQLA):
    """
    Just a simple test to verify that the TestMigrationsSQLA class indeed
    works and moves between the expected migration revisions
    """
    migrate_from = 'b8b23ddefad4'  # b8b23ddefad4_dbgroup_name_to_label_type_to_type_string.py
    migrate_to = 'e72ad251bcdb'  # e72ad251bcdb_dbgroup_class_change_type_string_values.py

    def setUpBeforeMigration(self):
        """
        Cache the start revision
        """
        self.start_revision = self.current_rev

    def test_revision_numbers(self):
        """
        Check that we went to the correct version
        """
        self.assertEqual(self.start_revision, self.migrate_from)
        self.assertEqual(self.current_rev, self.migrate_to)


class TestMigrationSchemaVsModelsSchema(AiidaTestCase):
    """
    This class checks that the schema that results from a migration is the
    same generated by the models. This is important since migrations are
    frequently written by hand or extended manually and we have to ensure
    that the final result is what is conceived in the SQLA models.
    """
    # The path to the folder that contains the migration configuration (the
    # actual configuration - not the testing)
    migr_method_dir_path = None
    # The path of the migration configuration (the actual configuration - not
    # the testing)
    alembic_dpath = None
    # The alembic configuration needed for the migrations is stored here
    alembic_cfg_left = None

    # The URL of the databases
    db_url_left = None
    db_url_right = None

    def setUp(self):
        from sqlalchemydiff.util import get_temporary_uri
        from aiida.backends.sqlalchemy.migrations import versions

        self.migr_method_dir_path = os.path.dirname(os.path.realpath(sqlalchemy_utils.__file__))
        # Set the alembic script directory location
        self.alembic_dpath = os.path.join(self.migr_method_dir_path, sqlalchemy_utils.ALEMBIC_REL_PATH)

        # Constructing the versions directory
        versions_dpath = os.path.join(os.path.dirname(versions.__file__))

        # Setting dynamically the the path to the alembic configuration
        # (this is where the env.py file can be found)
        self.alembic_cfg_left = Config()
        self.alembic_cfg_left.set_main_option('script_location', self.alembic_dpath)
        # Setting dynamically the versions directory. These are the
        # migration scripts to pass from one version to the other. The
        # default ones are overridden with test-specific migrations.
        self.alembic_cfg_left.set_main_option('version_locations', versions_dpath)

        # The correction URL to the SQLA database of the current
        # AiiDA connection
        curr_db_url = sa.engine.url

        # Create new urls for the two new databases
        self.db_url_left = get_temporary_uri(str(curr_db_url))
        self.db_url_right = get_temporary_uri(str(curr_db_url))

        # Put the correct database url to the database used by alembic
        self.alembic_cfg_left.set_main_option("sqlalchemy.url", self.db_url_left)

        # Database creation
        new_database(self.db_url_left)
        new_database(self.db_url_right)

    def tearDown(self):
        from sqlalchemydiff.util import destroy_database
        destroy_database(self.db_url_left)
        destroy_database(self.db_url_right)

    def test_model_and_migration_schemas_are_the_same(self):  # pylint: disable=invalid-name
        """Compare two databases.

        Compares the database obtained with all migrations against the
        one we get out of the models.  It produces a text file with the
        results to help debug differences.
        """
        from sqlalchemy.engine import create_engine  # pylint: disable=import-error,no-name-in-module
        from sqlalchemydiff import compare

        with create_engine(self.db_url_left).begin() as connection:
            self.alembic_cfg_left.attributes['connection'] = connection  # pylint: disable=unsupported-assignment-operation
            command.upgrade(self.alembic_cfg_left, "head")

        engine_right = create_engine(self.db_url_right)
        Base.metadata.create_all(engine_right)
        engine_right.dispose()

        result = compare(self.db_url_left, self.db_url_right, set(['alembic_version']))

        self.assertTrue(result.is_match, "The migration database doesn't match to the one "
                        "created by the models.\nDifferences: " + result._dump_data(result.errors))  # pylint: disable=protected-access


class TestProvenanceRedesignMigration(TestMigrationsSQLA):
    """Test the data migration part of the provenance redesign migration."""

    migrate_from = '140c971ae0a3'  # 140c971ae0a3_migrate_builtin_calculations
    migrate_to = '239cea6d2452'  # 239cea6d2452_provenance_redesign

    def setUpBeforeMigration(self):
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email='{}@aiida.net'.format(self.id()))
                session.add(user)
                session.commit()

                node_calc_job_known = DbNode(
                    type='calculation.job.arithmetic.add.ArithmeticAddCalculation.', user_id=user.id)
                node_calc_job_unknown = DbNode(type='calculation.job.unknown.PluginJobCalculation.', user_id=user.id)
                node_process = DbNode(type='calculation.process.ProcessCalculation.', user_id=user.id)
                node_work_chain = DbNode(type='calculation.work.WorkCalculation.', user_id=user.id)
                node_work_function = DbNode(
                    type='calculation.work.WorkCalculation.', attributes={'function_name': 'test'}, user_id=user.id)
                node_inline = DbNode(type='calculation.inline.InlineCalculation.', user_id=user.id)
                node_function = DbNode(type='calculation.function.FunctionCalculation.', user_id=user.id)

                session.add(node_calc_job_known)
                session.add(node_calc_job_unknown)
                session.add(node_process)
                session.add(node_work_chain)
                session.add(node_work_function)
                session.add(node_inline)
                session.add(node_function)
                session.commit()

                self.node_calc_job_known_id = node_calc_job_known.id
                self.node_calc_job_unknown_id = node_calc_job_unknown.id
                self.node_process_id = node_process.id
                self.node_work_chain_id = node_work_chain.id
                self.node_work_function_id = node_work_function.id
                self.node_inline_id = node_inline.id
                self.node_function_id = node_function.id
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def test_verify_migration(self):
        """Verify that type string of the Data node was successfully adapted."""
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                # Migration of calculation job with known plugin class
                node_calc_job_known = session.query(DbNode).filter(DbNode.id == self.node_calc_job_known_id).one()
                self.assertEqual(node_calc_job_known.type, 'node.process.calculation.calcjob.CalcJobNode.')
                self.assertEqual(node_calc_job_known.process_type, 'aiida.calculations:arithmetic.add')

                # Migration of calculation job with unknown plugin class
                node_calc_job_unknown = session.query(DbNode).filter(DbNode.id == self.node_calc_job_unknown_id).one()
                self.assertEqual(node_calc_job_unknown.type, 'node.process.calculation.calcjob.CalcJobNode.')
                self.assertEqual(node_calc_job_unknown.process_type, 'unknown.PluginJobCalculation')

                # Migration of very old `ProcessNode` class
                node_process = session.query(DbNode).filter(DbNode.id == self.node_process_id).one()
                self.assertEqual(node_process.type, 'node.process.workflow.workchain.WorkChainNode.')

                # Migration of old `WorkCalculation` class
                node_work_chain = session.query(DbNode).filter(DbNode.id == self.node_work_chain_id).one()
                self.assertEqual(node_work_chain.type, 'node.process.workflow.workchain.WorkChainNode.')

                # Migration of old `WorkCalculation` class used for work function
                node_work_function = session.query(DbNode).filter(DbNode.id == self.node_work_function_id).one()
                self.assertEqual(node_work_function.type, 'node.process.workflow.workfunction.WorkFunctionNode.')

                # Migration of old `InlineCalculation` class
                node_inline = session.query(DbNode).filter(DbNode.id == self.node_inline_id).one()
                self.assertEqual(node_inline.type, 'node.process.calculation.calcfunction.CalcFunctionNode.')

                # Migration of old `FunctionCalculation` class
                node_function = session.query(DbNode).filter(DbNode.id == self.node_function_id).one()
                self.assertEqual(node_function.type, 'node.process.workflow.workfunction.WorkFunctionNode.')

            finally:
                session.close()


class TestGroupRenamingMigration(TestMigrationsSQLA):
    """
    Test the migration that renames the DbGroup type strings
    """

    migrate_from = 'b8b23ddefad4'  # b8b23ddefad4_dbgroup_name_to_label_type_to_type_string.py
    migrate_to = 'e72ad251bcdb'  # e72ad251bcdb_dbgroup_class_change_type_string_values.py

    def setUpBeforeMigration(self):
        """
        Create the DbGroups with the old type strings
        """
        # Create group
        DbGroup = self.get_current_table('db_dbgroup')  # pylint: disable=invalid-name
        DbUser = self.get_current_table('db_dbuser')  # pylint: disable=invalid-name

        with self.get_session() as session:
            try:
                default_user = DbUser(is_superuser=False, email="{}@aiida.net".format(self.id()))
                session.add(default_user)
                session.commit()

                # test user group type_string: '' -> 'user'
                group_user = DbGroup(label='test_user_group', user_id=default_user.id, type_string='')
                session.add(group_user)
                # test data.upf group type_string: 'data.upf.family' -> 'data.upf'
                group_data_upf = DbGroup(
                    label='test_data_upf_group', user_id=default_user.id, type_string='data.upf.family')
                session.add(group_data_upf)
                # test auto.import group type_string: 'aiida.import' -> 'auto.import'
                group_autoimport = DbGroup(
                    label='test_import_group', user_id=default_user.id, type_string='aiida.import')
                session.add(group_autoimport)
                # test auto.run group type_string: 'autogroup.run' -> 'auto.run'
                group_autorun = DbGroup(
                    label='test_autorun_group', user_id=default_user.id, type_string='autogroup.run')
                session.add(group_autorun)

                session.commit()

                # Store values for later tests
                self.group_user_pk = group_user.id
                self.group_data_upf_pk = group_data_upf.id
                self.group_autoimport_pk = group_autoimport.id
                self.group_autorun_pk = group_autorun.id

            finally:
                session.close()

    def test_group_string_update(self):
        """
        Test that the type strings are properly migrated
        """
        DbGroup = self.get_current_table('db_dbgroup')  # pylint: disable=invalid-name

        with self.get_session() as session:
            try:
                # test user group type_string: '' -> 'user'
                group_user = session.query(DbGroup).filter(DbGroup.id == self.group_user_pk).one()
                self.assertEqual(group_user.type_string, 'user')

                # test data.upf group type_string: 'data.upf.family' -> 'data.upf'
                group_data_upf = session.query(DbGroup).filter(DbGroup.id == self.group_data_upf_pk).one()
                self.assertEqual(group_data_upf.type_string, 'data.upf')

                # test auto.import group type_string: 'aiida.import' -> 'auto.import'
                group_autoimport = session.query(DbGroup).filter(DbGroup.id == self.group_autoimport_pk).one()
                self.assertEqual(group_autoimport.type_string, 'auto.import')

                # test auto.run group type_string: 'autogroup.run' -> 'auto.run'
                group_autorun = session.query(DbGroup).filter(DbGroup.id == self.group_autorun_pk).one()
                self.assertEqual(group_autorun.type_string, 'auto.run')
            finally:
                session.close()


class TestCalcAttributeKeysMigration(TestMigrationsSQLA):
    """Test the migration of the keys of certain attribute for ProcessNodes and CalcJobNodes."""

    migrate_from = 'e72ad251bcdb'  # e72ad251bcdb_dbgroup_class_change_type_string_values
    migrate_to = '7ca08c391c49'  # 7ca08c391c49_calc_job_option_attribute_keys

    KEY_RESOURCES_OLD = 'jobresource_params'
    KEY_RESOURCES_NEW = 'resources'
    KEY_PARSER_NAME_OLD = 'parser'
    KEY_PARSER_NAME_NEW = 'parser_name'
    KEY_PROCESS_LABEL_OLD = '_process_label'
    KEY_PROCESS_LABEL_NEW = 'process_label'
    KEY_ENVIRONMENT_VARIABLES_OLD = 'custom_environment_variables'
    KEY_ENVIRONMENT_VARIABLES_NEW = 'environment_variables'

    def setUpBeforeMigration(self):
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email="{}@aiida.net".format(self.id()))
                session.add(user)
                session.commit()

                self.resources = {'number_machines': 1}
                self.parser_name = 'aiida.parsers:parser'
                self.process_label = 'TestLabel'
                self.environment_variables = {}

                attributes = {
                    self.KEY_RESOURCES_OLD: self.resources,
                    self.KEY_PARSER_NAME_OLD: self.parser_name,
                    self.KEY_PROCESS_LABEL_OLD: self.process_label,
                    self.KEY_ENVIRONMENT_VARIABLES_OLD: self.environment_variables,
                }
                node_work = DbNode(type='node.process.workflow.WorkflowNode.', attributes=attributes, user_id=user.id)
                node_calc = DbNode(
                    type='node.process.calculation.calcjob.CalcJobNode.', attributes=attributes, user_id=user.id)

                session.add(node_work)
                session.add(node_calc)
                session.commit()

                self.node_work_id = node_work.id
                self.node_calc_id = node_calc.id
            finally:
                session.close()

    def test_attribute_key_changes(self):
        """Verify that the keys are successfully changed of the affected attributes."""
        import ast
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name

        not_found = tuple([0])

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                node_work = session.query(DbNode).filter(DbNode.id == self.node_work_id).one()
                self.assertEqual(node_work.attributes.get(self.KEY_PROCESS_LABEL_NEW), self.process_label)
                self.assertEqual(node_work.attributes.get(self.KEY_PROCESS_LABEL_OLD, not_found), not_found)

                node_calc = session.query(DbNode).filter(DbNode.id == self.node_calc_id).one()

                # The dictionaries need to be cast with ast.literal_eval, because the `get` will return a string
                # representation of the dictionary that the attribute contains
                self.assertEqual(node_calc.attributes.get(self.KEY_PROCESS_LABEL_NEW), self.process_label)
                self.assertEqual(node_calc.attributes.get(self.KEY_PARSER_NAME_NEW), self.parser_name)
                self.assertEqual(ast.literal_eval(node_calc.attributes.get(self.KEY_RESOURCES_NEW)), self.resources)
                self.assertEqual(
                    ast.literal_eval(node_calc.attributes.get(self.KEY_ENVIRONMENT_VARIABLES_NEW)),
                    self.environment_variables)
                self.assertEqual(node_calc.attributes.get(self.KEY_PROCESS_LABEL_OLD, not_found), not_found)
                self.assertEqual(node_calc.attributes.get(self.KEY_PARSER_NAME_OLD, not_found), not_found)
                self.assertEqual(node_calc.attributes.get(self.KEY_RESOURCES_OLD, not_found), not_found)
                self.assertEqual(node_calc.attributes.get(self.KEY_ENVIRONMENT_VARIABLES_OLD, not_found), not_found)
            finally:
                session.close()


class TestDbLogMigrationRecordCleaning(TestMigrationsSQLA):
    """Test the migration of the keys of certain attribute for ProcessNodes and CalcJobNodes."""

    migrate_from = '7ca08c391c49'  # 7ca08c391c49_calc_job_option_attribute_keys
    migrate_to = '041a79fc615f'  # 041a79fc615f_dblog_cleaning

    def setUpBeforeMigration(self):
        # pylint: disable=too-many-locals,too-many-statements
        import importlib
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module
        from aiida.backends.sqlalchemy.utils import dumps_json

        log_migration = importlib.import_module(
            'aiida.backends.sqlalchemy.migrations.versions.041a79fc615f_dblog_cleaning')

        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name
        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbWorkflow = self.get_auto_base().classes.db_dbworkflow  # pylint: disable=invalid-name
        DbLog = self.get_auto_base().classes.db_dblog  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email="{}@aiida.net".format(self.id()))
                session.add(user)
                session.commit()

                calc_1 = DbNode(type="node.process.calculation.CalculationNode.", user_id=user.id)
                param = DbNode(type="data.dict.Dict.", user_id=user.id)
                leg_workf = DbWorkflow(label="Legacy WorkflowNode", user_id=user.id)
                calc_2 = DbNode(type="node.process.calculation.CalculationNode.", user_id=user.id)

                session.add(calc_1)
                session.add(param)
                session.add(leg_workf)
                session.add(calc_2)
                session.commit()

                log_1 = DbLog(
                    loggername='CalculationNode logger',
                    objpk=calc_1.id,
                    objname='node.calculation.job.quantumespresso.pw.',
                    message='calculation node 1',
                    metadata={
                        "msecs": 719.0849781036377,
                        "objpk": calc_1.id,
                        "lineno": 350,
                        "thread": 140011612940032,
                        "asctime": "10/21/2018 12:39:51 PM",
                        "created": 1540118391.719085,
                        "levelno": 23,
                        "message": "calculation node 1",
                        "objname": "node.calculation.job.quantumespresso.pw.",
                    })
                log_2 = DbLog(
                    loggername='something.else logger',
                    objpk=param.id,
                    objname='something.else.',
                    message='parameter data with log message')
                log_3 = DbLog(
                    loggername='TopologicalWorkflow logger',
                    objpk=leg_workf.id,
                    objname='aiida.workflows.user.topologicalworkflows.topo.TopologicalWorkflow',
                    message='parameter data with log message')
                log_4 = DbLog(
                    loggername='CalculationNode logger',
                    objpk=calc_2.id,
                    objname='node.calculation.job.quantumespresso.pw.',
                    message='calculation node 2',
                    metadata={
                        "msecs": 719.0849781036377,
                        "objpk": calc_2.id,
                        "lineno": 360,
                        "levelno": 23,
                        "message": "calculation node 1",
                        "objname": "node.calculation.job.quantumespresso.pw.",
                    })
                # Creating two more log records that don't correspond to a node
                log_5 = DbLog(
                    loggername='CalculationNode logger',
                    objpk=(calc_2.id + 1000),
                    objname='node.calculation.job.quantumespresso.pw.',
                    message='calculation node 1000',
                    metadata={
                        "msecs": 718,
                        "objpk": (calc_2.id + 1000),
                        "lineno": 361,
                        "levelno": 25,
                        "message": "calculation node 1000",
                        "objname": "node.calculation.job.quantumespresso.pw.",
                    })
                log_6 = DbLog(
                    loggername='CalculationNode logger',
                    objpk=(calc_2.id + 1001),
                    objname='node.calculation.job.quantumespresso.pw.',
                    message='calculation node 10001',
                    metadata={
                        "msecs": 722,
                        "objpk": (calc_2.id + 1001),
                        "lineno": 362,
                        "levelno": 24,
                        "message": "calculation node 1001",
                        "objname": "node.calculation.job.quantumespresso.pw.",
                    })

                session.add(log_1)
                session.add(log_2)
                session.add(log_3)
                session.add(log_4)
                session.add(log_5)
                session.add(log_6)

                session.commit()

                # Storing temporarily information needed for the check at the test
                self.to_check = dict()

                # Keeping calculation & calculation log ids
                self.to_check['CalculationNode'] = (
                    calc_1.id,
                    log_1.id,
                    calc_2.id,
                    log_4.id,
                )

                # The columns to project
                cols_to_project = []
                for val in log_migration.values_to_export:
                    cols_to_project.append(getattr(DbLog, val))

                # Getting the serialized Dict logs
                param_data = session.query(DbLog).filter(DbLog.objpk == param.id).filter(
                    DbLog.objname == 'something.else.').with_entities(*cols_to_project).one()
                serialized_param_data = dumps_json([(dict(list(zip(param_data.keys(), param_data))))])
                # Getting the serialized logs for the unknown entity logs (as the export migration fuction
                # provides them) - this should coincide to the above
                serialized_unknown_exp_logs = log_migration.get_serialized_unknown_entity_logs(connection)
                # Getting their number
                unknown_exp_logs_number = log_migration.get_unknown_entity_log_number(connection)
                self.to_check['Dict'] = (serialized_param_data, serialized_unknown_exp_logs, unknown_exp_logs_number)

                # Getting the serialized legacy workflow logs
                # yapf: disable
                leg_wf = session.query(DbLog).filter(DbLog.objpk == leg_workf.id).filter(
                    DbLog.objname == 'aiida.workflows.user.topologicalworkflows.topo.TopologicalWorkflow'
                ).with_entities(*cols_to_project).one()
                serialized_leg_wf_logs = dumps_json([(dict(list(zip(leg_wf.keys(), leg_wf))))])
                # Getting the serialized logs for the legacy workflow logs (as the export migration function
                # provides them) - this should coincide to the above
                serialized_leg_wf_exp_logs = log_migration.get_serialized_legacy_workflow_logs(connection)
                eg_wf_exp_logs_number = log_migration.get_legacy_workflow_log_number(connection)
                self.to_check['WorkflowNode'] = (serialized_leg_wf_logs, serialized_leg_wf_exp_logs,
                                                 eg_wf_exp_logs_number)

                # Getting the serialized logs that don't correspond to a DbNode record
                logs_no_node = session.query(DbLog).filter(
                    DbLog.id.in_([log_5.id, log_6.id])).with_entities(*cols_to_project)
                logs_no_node_list = list()
                for log_no_node in logs_no_node:
                    logs_no_node_list.append((dict(list(zip(log_no_node.keys(), log_no_node)))))
                serialized_logs_no_node = dumps_json(logs_no_node_list)

                # Getting the serialized logs that don't correspond to a node (as the export migration function
                # provides them) - this should coincide to the above
                serialized_logs_exp_no_node = log_migration.get_serialized_logs_with_no_nodes(connection)
                logs_no_node_number = log_migration.get_logs_with_no_nodes_number(connection)
                self.to_check['NoNode'] = (serialized_logs_no_node, serialized_logs_exp_no_node, logs_no_node_number)
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def test_dblog_calculation_node(self):
        """
        Verify that after the migration there is only two log records left and verify that they corresponds to
        the CalculationNodes.
        """
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbLog = self.get_auto_base().classes.db_dblog  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                # Check that only two log records exist
                self.assertEqual(session.query(DbLog).count(), 2, "There should be two log records left")

                # Get the node id of the log record referencing the node and verify that it is the correct one
                dbnode_id_1 = session.query(DbLog).filter(
                    DbLog.id == self.to_check['CalculationNode'][1]).with_entities('dbnode_id').one()[0]
                self.assertEqual(dbnode_id_1, self.to_check['CalculationNode'][0], "The the referenced node is not "
                                 "the expected one")
                dbnode_id_2 = session.query(DbLog).filter(
                    DbLog.id == self.to_check['CalculationNode'][3]).with_entities('dbnode_id').one()[0]
                self.assertEqual(dbnode_id_2, self.to_check['CalculationNode'][2], "The the referenced node is not "
                                 "the expected one")
            finally:
                session.close()

    def test_dblog_correct_export_of_logs(self):
        """
        Verify that export log methods for legacy workflows, unknown entities and log records that
        don't correspond to nodes, work as expected
        """
        import json

        self.assertEqual(self.to_check['Dict'][0], self.to_check['Dict'][1])
        self.assertEqual(self.to_check['Dict'][2], 1)

        self.assertEqual(self.to_check['WorkflowNode'][0], self.to_check['WorkflowNode'][1])
        self.assertEqual(self.to_check['WorkflowNode'][2], 1)

        self.assertEqual(sorted(list(json.loads(self.to_check['NoNode'][0])), key=lambda k: k['id']),
                          sorted(list(json.loads(self.to_check['NoNode'][1])), key=lambda k: k['id']))
        self.assertEqual(self.to_check['NoNode'][2], 2)

    def test_metadata_correctness(self):
        """
        Verify that the metadata of the remaining records don't have an objpk and objmetadata values.
        """
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbLog = self.get_auto_base().classes.db_dblog  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)
                metadata = list(session.query(DbLog).with_entities(getattr(DbLog, 'metadata')).all())
                # Verify that the objpk and objname are no longer part of the metadata
                for (m_res,) in metadata:
                    self.assertNotIn('objpk', m_res.keys(), "objpk should not exist any more in metadata")
                    self.assertNotIn('objname', m_res.keys(), "objname should not exist any more in metadata")

            finally:
                session.close()


class TestDbLogMigrationBackward(TestBackwardMigrationsSQLA):
    """
    Check that backward migrations work also for the DbLog migration(s).
    """
    migrate_from = '041a79fc615f'  # 041a79fc615f_dblog_cleaning
    migrate_to = '7ca08c391c49'  # e72ad251bcdb_dbgroup_class_change_type_string_values

    def setUpBeforeMigration(self):
        # pylint: disable=too-many-locals,too-many-statements
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name
        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbLog = self.get_auto_base().classes.db_dblog  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email="{}@aiida.net".format(self.id()))
                session.add(user)
                session.commit()

                calc_1 = DbNode(type="node.process.calculation.CalculationNode.1", user_id=user.id)
                calc_2 = DbNode(type="node.process.calculation.CalculationNode.2", user_id=user.id)

                session.add(calc_1)
                session.add(calc_2)
                session.commit()

                log_1 = DbLog(
                    loggername='CalculationNode logger',
                    dbnode_id=calc_1.id,
                    message='calculation node 1',
                    metadata={
                        "msecs": 719.0849781036377,
                        "lineno": 350,
                        "thread": 140011612940032,
                        "asctime": "10/21/2018 12:39:51 PM",
                        "created": 1540118391.719085,
                        "levelno": 23,
                        "message": "calculation node 1",
                    })
                log_2 = DbLog(
                    loggername='CalculationNode logger',
                    dbnode_id=calc_2.id,
                    message='calculation node 2',
                    metadata={
                        "msecs": 719.0849781036377,
                        "lineno": 360,
                        "levelno": 23,
                        "message": "calculation node 1",
                    })

                session.add(log_1)
                session.add(log_2)

                session.commit()

                # Keeping what is needed to be verified at the test
                self.to_check = dict()
                self.to_check[log_1.id] = (log_1.dbnode_id, calc_1.type)
                self.to_check[log_2.id] = (log_2.dbnode_id, calc_2.type)
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def test_objpk_objname(self):
        """
        This test verifies that the objpk and objname have the right values
        after a forward and a backward migration.
        """
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbLog = self.get_auto_base().classes.db_dblog  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                for log_pk in self.to_check:
                    log_entry = session.query(DbLog).filter(DbLog.id == log_pk).one()
                    log_dbnode_id, node_type = self.to_check[log_pk]

                    self.assertEqual(
                        log_dbnode_id, log_entry.objpk,
                        "The dbnode_id ({}) of the 0024 schema version should be identical to the objpk ({}) of "
                        "the 0023 schema version.".format(log_dbnode_id, log_entry.objpk))
                    self.assertEqual(
                        node_type, log_entry.objname,
                        "The type ({}) of the linked node of the 0024 schema version should be identical to the "
                        "objname ({}) of the 0023 schema version.".format(node_type, log_entry.objname))
                    self.assertEqual(
                        log_dbnode_id, log_entry.metadata['objpk'],
                        "The dbnode_id ({}) of the 0024 schema version should be identical to the objpk ({}) of "
                        "the 0023 schema version stored in the metadata.".format(log_dbnode_id,
                                                                                 log_entry.metadata['objpk']))
                    self.assertEqual(
                        node_type, log_entry.metadata['objname'],
                        "The type ({}) of the linked node of the 0024 schema version should be identical to the "
                        "objname ({}) of the 0023 schema version stored in the metadata.".format(
                            node_type, log_entry.metadata['objname']))
            finally:
                session.close()


class TestDbLogUUIDAddition(TestMigrationsSQLA):
    """
    Test that the UUID column is correctly added to the DbLog table and that the uniqueness
    constraint is added without problems (if the migration arrives until 375c2db70663 then the
    constraint is added properly.
    """

    migrate_from = '041a79fc615f'  # 041a79fc615f_dblog_cleaning
    migrate_to = '375c2db70663'  # 375c2db70663_dblog_uuid_uniqueness_constraint

    def setUpBeforeMigration(self):
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name
        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbLog = self.get_auto_base().classes.db_dblog  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email="{}@aiida.net".format(self.id()))
                session.add(user)
                session.commit()

                calc_1 = DbNode(type="node.process.calculation.CalculationNode.", user_id=user.id)
                calc_2 = DbNode(type="node.process.calculation.CalculationNode.", user_id=user.id)

                session.add(calc_1)
                session.add(calc_2)
                session.commit()

                log_1 = DbLog(loggername='CalculationNode logger', dbnode_id=calc_1.id, message='calculation node 1')
                log_2 = DbLog(loggername='CalculationNode logger', dbnode_id=calc_2.id, message='calculation node 2')

                session.add(log_1)
                session.add(log_2)

                session.commit()
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def test_dblog_unique_uuids(self):
        """
        Verify that the UUIDs of the log records are unique
        """
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbLog = self.get_auto_base().classes.db_dblog  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)
                l_uuids = list(session.query(DbLog).with_entities(getattr(DbLog, 'uuid')).all())
                s_uuids = set(l_uuids)
                self.assertEqual(len(l_uuids), len(s_uuids), "The UUIDs are not all unique.")
            finally:
                session.close()


class TestDataMoveWithinNodeMigration(TestMigrationsSQLA):
    """Test the migration of Data nodes after the data module was moved within the node moduel."""

    migrate_from = '041a79fc615f'  # 041a79fc615f_dblog_update
    migrate_to = '6a5c2ea1439d'  # 6a5c2ea1439d_move_data_within_node_module

    def setUpBeforeMigration(self):
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email="{}@aiida.net".format(self.id()))
                session.add(user)
                session.commit()

                node_calc = DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=user.id)
                node_data = DbNode(type='data.int.Int.', user_id=user.id)

                session.add(node_data)
                session.add(node_calc)
                session.commit()

                self.node_calc_id = node_calc.id
                self.node_data_id = node_data.id
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def test_data_node_type_string(self):
        """Verify that type string of the Data node was successfully adapted."""
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                # The data node should have been touched and migrated
                node_data = session.query(DbNode).filter(DbNode.id == self.node_data_id).one()
                self.assertEqual(node_data.type, 'node.data.int.Int.')

                # The calc node by contrast should not have been changed
                node_calc = session.query(DbNode).filter(DbNode.id == self.node_calc_id).one()
                self.assertEqual(node_calc.type, 'node.process.calculation.calcjob.CalcJobNode.')
            finally:
                session.close()


class TestTrajectoryDataMigration(TestMigrationsSQLA):
    """Test the migration of the symbols from numpy array to attribute for TrajectoryData nodes."""
    import numpy

    migrate_from = '37f3d4882837'  # 37f3d4882837_make_all_uuid_columns_unique
    migrate_to = 'ce56d84bcc35'  # ce56d84bcc35_delete_trajectory_symbols_array

    stepids = numpy.array([60, 70])
    times = stepids * 0.01
    positions = numpy.array(
        [[[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]], [[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]]])
    velocities = numpy.array(
        [[[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]], [[0.5, 0.5, 0.5], [0.5, 0.5, 0.5], [-0.5, -0.5, -0.5]]])
    cells = numpy.array([[[2., 0., 0.], [0., 2., 0.], [0., 0., 2.]], [[3., 0., 0.], [0., 3., 0.], [0., 0., 3.]]])

    def setUpBeforeMigration(self):
        import numpy
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email='{}@aiida.net'.format(self.id()))
                session.add(user)
                session.commit()

                node = DbNode(type='node.data.array.trajectory.TrajectoryData.', user_id=user.id)
                session.add(node)
                session.commit()

                symbols = numpy.array(['H', 'O', 'C'])

                self.set_node_array(node, 'steps', self.stepids)
                self.set_node_array(node, 'cells', self.cells)
                self.set_node_array(node, 'symbols', symbols)
                self.set_node_array(node, 'positions', self.positions)
                self.set_node_array(node, 'times', self.times)
                self.set_node_array(node, 'velocities', self.velocities)
                session.commit()

                self.node_uuid = node.uuid
            finally:
                session.close()

    def test_trajectory_symbols(self):
        """Verify that migration of symbols from repository array to attribute works properly."""
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                node = session.query(DbNode).filter(DbNode.uuid == self.node_uuid).one()

                self.assertSequenceEqual(node.attributes['symbols'], ['H', 'O', 'C'])
                self.assertSequenceEqual(self.get_node_array(node, 'velocities').tolist(), self.velocities.tolist())
                self.assertSequenceEqual(self.get_node_array(node, 'positions').tolist(), self.positions.tolist())
                with self.assertRaises(IOError):
                    self.get_node_array(node, 'symbols')

            finally:
                session.close()


class TestNodePrefixRemovalMigration(TestMigrationsSQLA):
    """Test the migration of Data nodes after the data module was moved within the node moduel."""

    migrate_from = 'ce56d84bcc35'  # ce56d84bcc35_delete_trajectory_symbols_array
    migrate_to = '61fc0913fae9'  # 61fc0913fae9_remove_node_prefix

    def setUpBeforeMigration(self):
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email='{}@aiida.net'.format(self.id()))
                session.add(user)
                session.commit()

                node_calc = DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=user.id)
                node_data = DbNode(type='node.data.int.Int.', user_id=user.id)

                session.add(node_data)
                session.add(node_calc)
                session.commit()

                self.node_calc_id = node_calc.id
                self.node_data_id = node_data.id
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def test_data_node_type_string(self):
        """Verify that type string of the Data node was successfully adapted."""
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                # Verify that the `node.` prefix has been dropped from both the data as well as the process node
                node_data = session.query(DbNode).filter(DbNode.id == self.node_data_id).one()
                self.assertEqual(node_data.type, 'data.int.Int.')

                node_calc = session.query(DbNode).filter(DbNode.id == self.node_calc_id).one()
                self.assertEqual(node_calc.type, 'process.calculation.calcjob.CalcJobNode.')
            finally:
                session.close()


class TestParameterDataToDictMigration(TestMigrationsSQLA):
    """Test the data migration after `ParameterData` was renamed to `Dict`."""

    migrate_from = '61fc0913fae9'  # 61fc0913fae9_remove_node_prefix
    migrate_to = 'd254fdfed416'  # d254fdfed416_rename_parameter_data_to_dict

    def setUpBeforeMigration(self):
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name
        DbUser = self.get_auto_base().classes.db_dbuser  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                user = DbUser(is_superuser=False, email='{}@aiida.net'.format(self.id()))
                session.add(user)
                session.commit()

                node = DbNode(type='data.parameter.ParameterData.', user_id=user.id)

                session.add(node)
                session.commit()

                self.node_id = node.id
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def test_type_string(self):
        """Verify that type string of the Data node was successfully adapted."""
        from sqlalchemy.orm import Session  # pylint: disable=import-error,no-name-in-module

        DbNode = self.get_auto_base().classes.db_dbnode  # pylint: disable=invalid-name

        with sa.engine.begin() as connection:
            try:
                session = Session(connection.engine)

                node = session.query(DbNode).filter(DbNode.id == self.node_id).one()
                self.assertEqual(node.type, 'data.dict.Dict.')
            finally:
                session.close()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the `ProcessSpec` class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.engine import Process
from aiida.orm import Node, Data


class TestProcessSpec(AiidaTestCase):
    """Tests for the `ProcessSpec` class."""

    def setUp(self):
        super(TestProcessSpec, self).setUp()
        self.assertIsNone(Process.current())
        self.spec = Process.spec()

    def tearDown(self):
        super(TestProcessSpec, self).tearDown()
        self.assertIsNone(Process.current())

    def test_dynamic_input(self):
        """Test a process spec with dynamic input enabled."""
        node = Node()
        data = Data()
        self.assertIsNotNone(self.spec.validate_inputs({'key': 'foo'}))
        self.assertIsNotNone(self.spec.validate_inputs({'key': 5}))
        self.assertIsNotNone(self.spec.validate_inputs({'key': node}))
        self.assertIsNone(self.spec.validate_inputs({'key': data}))

    def test_dynamic_output(self):
        """Test a process spec with dynamic output enabled."""
        node = Node()
        data = Data()
        self.assertIsNotNone(self.spec.validate_outputs({'key': 'foo'}))
        self.assertIsNotNone(self.spec.validate_outputs({'key': 5}))
        self.assertIsNotNone(self.spec.validate_outputs({'key': node}))
        self.assertIsNone(self.spec.validate_outputs({'key': data}))

    def test_exit_code(self):
        """Test the definition of error codes through the ProcessSpec."""
        label = 'SOME_EXIT_CODE'
        status = 418
        message = 'I am a teapot'

        self.spec.exit_code(status, label, message)

        self.assertEqual(self.spec.exit_codes.SOME_EXIT_CODE.status, status)
        self.assertEqual(self.spec.exit_codes.SOME_EXIT_CODE.message, message)

        self.assertEqual(self.spec.exit_codes['SOME_EXIT_CODE'].status, status)
        self.assertEqual(self.spec.exit_codes['SOME_EXIT_CODE'].message, message)

        self.assertEqual(self.spec.exit_codes[label].status, status)
        self.assertEqual(self.spec.exit_codes[label].message, message)

    def test_exit_code_invalid(self):
        """Test type validation for registering new error codes."""
        status = 418
        label = 'SOME_EXIT_CODE'
        message = 'I am a teapot'

        with self.assertRaises(TypeError):
            self.spec.exit_code(status, 256, message)

        with self.assertRaises(TypeError):
            self.spec.exit_code('string', label, message)

        with self.assertRaises(ValueError):
            self.spec.exit_code(-256, label, message)

        with self.assertRaises(TypeError):
            self.spec.exit_code(status, label, 8)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi rehash` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments
from aiida.cmdline.params.types.plugin import PluginParamType
from aiida.cmdline.utils import decorators, echo


@verdi.command('rehash')
@arguments.NODES()
@click.option(
    '-e',
    '--entry-point',
    type=PluginParamType(group=('aiida.calculations', 'aiida.data', 'aiida.workflows'), load=True),
    default=None,
    help='Only include nodes that are class or sub class of the class identified by this entry point.')
@decorators.with_dbenv()
def rehash(nodes, entry_point):
    """Recompute the hash for nodes in the database

    The set of nodes that will be rehashed can be filtered by their identifier and/or based on their class.
    """
    from aiida.orm import Data, ProcessNode, QueryBuilder

    # If no explicit entry point is defined, rehash all nodes, which are either Data nodes or ProcessNodes
    if entry_point is None:
        entry_point = (Data, ProcessNode)

    if nodes:
        to_hash = [(node,) for node in nodes if isinstance(node, entry_point)]
    else:
        builder = QueryBuilder()
        builder.append(entry_point, tag='node')
        to_hash = builder.all()

    if not to_hash:
        echo.echo_critical('no matching nodes found')

    count = 0

    for i, (node,) in enumerate(to_hash):

        if i % 100 == 0:
            echo.echo('.', nl=False)

        node.rehash()
        count += 1

    echo.echo('')
    echo.echo_success('{} nodes re-hashed'.format(count))
